# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""Projects main widget."""

# Standard library imports
from collections import OrderedDict
import configparser
from contextlib import contextmanager
import logging
import os
import os.path as osp
import pathlib
import shutil

# Third party imports
from qtpy.compat import getexistingdirectory
from qtpy.QtCore import Qt, Signal, Slot
from qtpy.QtWidgets import (
    QHBoxLayout, QInputDialog, QLabel, QMessageBox, QVBoxLayout, QWidget)

# Local imports
from spyder.api.config.decorators import on_conf_change
from spyder.api.exceptions import SpyderAPIError
from spyder.api.translations import _
from spyder.api.widgets.main_widget import PluginMainWidget
from spyder.config.base import (
    get_home_dir, get_project_config_folder, running_under_pytest)
from spyder.config.utils import EDIT_EXTENSIONS
from spyder.plugins.completion.api import (
    CompletionRequestTypes, FileChangeType)
from spyder.plugins.completion.decorators import (
    class_register, handles, request)
from spyder.plugins.explorer.api import DirViewActions
from spyder.plugins.projects.api import (
    BaseProjectType, EmptyProject, WORKSPACE)
from spyder.plugins.projects.utils.watcher import WorkspaceWatcher
from spyder.plugins.projects.widgets.projectdialog import ProjectDialog
from spyder.plugins.projects.widgets.projectexplorer import (
    ProjectExplorerTreeWidget)
from spyder.plugins.switcher.utils import get_file_icon, shorten_paths
from spyder.utils import encoding
from spyder.utils.misc import getcwd_or_home
from spyder.utils.programs import find_program
from spyder.utils.workers import WorkerManager
from spyder.widgets.helperwidgets import PaneEmptyWidget


# For logging
logger = logging.getLogger(__name__)


# ---- Constants
# -----------------------------------------------------------------------------
class ProjectExplorerOptionsMenuSections:
    Main = 'main'


class ProjectsActions:
    NewProject = 'new_project_action'
    OpenProject = 'open_project_action'
    CloseProject = 'close_project_action'
    DeleteProject = 'delete_project_action'
    ClearRecentProjects = 'clear_recent_projects_action'
    MaxRecent = 'max_recent_action'


class ProjectsMenuSubmenus:
    RecentProjects = 'recent_projects'


class RecentProjectsMenuSections:
    Recent = 'recent_section'
    Extras = 'extras_section'


class ProjectsOptionsMenuActions:
    SearchInSwitcher = "search_in_switcher"


# ---- Main widget
# -----------------------------------------------------------------------------
@class_register
class ProjectExplorerWidget(PluginMainWidget):
    """Project explorer main widget."""

    # ---- Constants
    # -------------------------------------------------------------------------
    MAX_SWITCHER_RESULTS = 50

    # ---- Signals
    # -------------------------------------------------------------------------
    sig_open_file_requested = Signal(str)
    """
    This signal is emitted when a file is requested to be opened.

    Parameters
    ----------
    directory: str
        The path to the requested file.
    """

    sig_project_created = Signal(str, str)
    """
    This signal is emitted to request the Projects plugin the creation of a
    project.

    Parameters
    ----------
    project_path: str
        Location of project.
    project_type: str
        Type of project as defined by project types.
    """

    sig_project_loaded = Signal(str)
    """
    This signal is emitted when a project is loaded.

    Parameters
    ----------
    project_path: str
        Loaded project path.
    """

    sig_project_closed = Signal((str,), (bool,))
    """
    This signal is emitted when a project is closed.

    Parameters
    ----------
    project_path: str
        Closed project path (signature 1).
    close_project: bool
        This is emitted only when closing a project but not when switching
        between projects (signature 2).
    """

    sig_save_open_files_requested = Signal()
    """
    This signal is emitted to request saving the list of open files in the
    editor.
    """

    sig_restart_console_requested = Signal()
    """This signal is emitted to request restarting the IPython console."""

    sig_broadcast_notification_requested = Signal(str, dict)
    """
    This signal is emitted to request that the Completions plugin broadcast
    a notification.

    Parameters
    ----------
    method: str
        Method name to broadcast.
    params: dict
        Parameters of the notification.
    """

    def __init__(self, name, plugin, parent=None):
        super().__init__(name, plugin=plugin, parent=parent)

        # -- Attributes from conf
        self.name_filters = self.get_conf('name_filters')
        self.show_hscrollbar = self.get_conf('show_hscrollbar')

        # -- Main attributes
        self.recent_projects = self._get_valid_recent_projects(
            self.get_conf('recent_projects', [])
        )
        self._project_types = OrderedDict()
        self.current_active_project = None
        self.latest_project = None
        self.completions_available = False
        self._fzf = find_program('fzf')
        self._default_switcher_paths = []

        # -- Tree widget
        self.treewidget = ProjectExplorerTreeWidget(self, self.show_hscrollbar)
        self.treewidget.setup()
        self.treewidget.setup_view()
        self.treewidget.hide()
        self.treewidget.sig_open_file_requested.connect(
            self.sig_open_file_requested)

        self.pane_empty = PaneEmptyWidget(
            self,
            "projects",
            _("No project opened"),
            _("Create one using the menu entry Projects > New project.")
        )

        # -- Watcher
        self.watcher = WorkspaceWatcher(self)
        self.watcher.connect_signals(self)

        # -- Worker manager for calls to fzf
        self._worker_manager = WorkerManager(self)

        # -- Signals
        self.sig_project_loaded.connect(self._setup_project)

        # This is necessary to populate the switcher with some default list of
        # paths instead of computing that list every time it's shown.
        self.sig_project_loaded.connect(
            lambda p: self._update_default_switcher_paths()
        )

        # Clear saved paths for the switcher when closing the project.
        self.sig_project_closed.connect(lambda p: self._clear_switcher_paths())

        # -- Layout
        layout = QVBoxLayout()
        layout.addWidget(self.pane_empty)
        layout.addWidget(self.treewidget)
        self.setLayout(layout)
        self.setMinimumWidth(200)

        # Initial setup
        self._setup_project(self.get_active_project_path())

    # ---- PluginMainWidget API
    # -------------------------------------------------------------------------
    def get_title(self):
        return _("Project")

    def setup(self):
        """Setup the widget."""
        # Create default actions
        self.create_action(
            ProjectsActions.NewProject,
            text=_("New Project..."),
            triggered=self.create_new_project,
            icon=self.create_icon("project_new"))

        self.create_action(
            ProjectsActions.OpenProject,
            text=_("Open Project..."),
            triggered=lambda v: self.open_project(),
            icon=self.create_icon("project_open"))

        self.close_project_action = self.create_action(
            ProjectsActions.CloseProject,
            text=_("Close Project"),
            triggered=self.close_project,
            icon=self.create_icon("project_close"))

        self.delete_project_action = self.create_action(
            ProjectsActions.DeleteProject,
            text=_("Delete Project"),
            triggered=self.delete_project,
            icon=self.create_icon("project_delete"))

        self.clear_recent_projects_action = self.create_action(
            ProjectsActions.ClearRecentProjects,
            text=_("Clear this list"),
            triggered=self.clear_recent_projects)

        self.max_recent_action = self.create_action(
            ProjectsActions.MaxRecent,
            text=_("Maximum number of recent projects..."),
            icon=self.create_icon("transparent"),
            triggered=self.change_max_recent_projects)

        self.recent_project_menu = self.create_menu(
            ProjectsMenuSubmenus.RecentProjects,
            _("Recent Projects"),
            reposition=False
        )
        self.recent_project_menu.aboutToShow.connect(self._setup_menu_actions)
        self._setup_menu_actions()

        # We need to give users a way to disable searching files in the
        # switcher because in some situations it introduces delays in the
        # switcher or Spyder itself.
        # Fixes spyder-ide/spyder#22641
        search_in_switcher_action = self.create_action(
            ProjectsOptionsMenuActions.SearchInSwitcher,
            text=_("Search project files in the switcher"),
            toggled=True,
            option='search_files_in_switcher',
        )

        # Add some DirView actions to the Options menu for easy access.
        hidden_action = self.get_action(DirViewActions.ToggleHiddenFiles)
        single_click_action = self.get_action(DirViewActions.ToggleSingleClick)

        # Options menu
        menu = self.get_options_menu()
        for action in [
            hidden_action,
            single_click_action,
            search_in_switcher_action,
        ]:
            self.add_item_to_menu(
                action,
                menu=menu,
                section=ProjectExplorerOptionsMenuSections.Main
            )

    def set_pane_empty(self):
        self.treewidget.hide()
        self.pane_empty.show()

    def update_actions(self):
        pass

    def on_close(self):
        self._worker_manager.terminate_all()

    # ---- Public API
    # -------------------------------------------------------------------------
    @Slot()
    def create_new_project(self):
        """Create new project."""
        self._unmaximize()

        dlg = ProjectDialog(self, project_types=self.get_project_types())
        result = dlg.exec_()
        data = dlg.project_data
        root_path = data.get("root_path", None)
        project_type = data.get("project_type", EmptyProject.ID)

        if result:
            logger.debug(f'Creating a project at {root_path}')
            self.create_project(root_path, project_type_id=project_type)
            dlg.close()

    def create_project(self, root_path, project_type_id=EmptyProject.ID):
        """Create a new project."""
        project_types = self.get_project_types()
        if project_type_id in project_types:
            project_type_class = project_types[project_type_id]
            project_type = project_type_class(
                root_path=root_path,
                parent_plugin=project_type_class._PARENT_PLUGIN,
            )

            created_succesfully, message = project_type.create_project()
            if not created_succesfully:
                QMessageBox.warning(self, "Project creation", message)
                shutil.rmtree(root_path, ignore_errors=True)
                return

            self.sig_project_created.emit(root_path, project_type_id)
            self.open_project(path=root_path, project_type=project_type)
        else:
            if not running_under_pytest():
                QMessageBox.critical(
                    self,
                    _('Error'),
                    _("<b>{}</b> is not a registered Spyder project "
                      "type!").format(project_type_id)
                )

    def open_project(self, path=None, project_type=None, restart_console=True,
                     save_previous_files=True, workdir=None):
        """Open the project located in `path`."""
        self._unmaximize()

        if path is None:
            basedir = get_home_dir()
            path = getexistingdirectory(
                parent=self,
                caption=_("Open project"),
                basedir=basedir
            )

            path = encoding.to_unicode_from_fs(path)
            if not self.is_valid_project(path):
                if path:
                    QMessageBox.critical(
                        self,
                        _('Error'),
                        _("<b>%s</b> is not a Spyder project!") % path,
                    )
                return
        else:
            path = encoding.to_unicode_from_fs(path)

        # This makes the path have a uniform representation in all OSes. For
        # instance, it always uses backslashes as separators on Windows.
        path = str(pathlib.Path(path))

        logger.debug(f'Opening project located at {path}')

        if project_type is None:
            project_type_class = self._load_project_type_class(path)
            project_type = project_type_class(
                root_path=path,
                parent_plugin=project_type_class._PARENT_PLUGIN,
            )

        if self.current_active_project is None:
            # A project was not open before
            if save_previous_files:
                self.sig_save_open_files_requested.emit()

            self.set_conf('last_working_dir', getcwd_or_home(),
                          section='editor')

            if self.get_conf('visible_if_project_open'):
                self.show_widget()
        else:
            # We are switching projects
            filenames = self.get_plugin()._get_open_filenames()
            if filenames is not None:
                self.set_project_filenames(filenames)

            # TODO: Don't emit sig_project_closed when we support
            # multiple workspaces.
            self.sig_project_closed.emit(self.current_active_project.root_path)
            self.watcher.stop()

        self.current_active_project = project_type
        self.latest_project = project_type
        self._add_to_recent(path)

        self.set_conf('current_project_path', self.get_active_project_path())
        self._setup_menu_actions()

        with self._disable_pdb_prevent_closing():
            if workdir and osp.isdir(workdir):
                self.sig_project_loaded.emit(workdir)
            else:
                self.sig_project_loaded.emit(path)

        self.watcher.start(path)

        if restart_console:
            self.sig_restart_console_requested.emit()

        open_successfully, message = project_type.open_project()
        if not open_successfully:
            QMessageBox.warning(self, "Project open", message)

    def close_project(self):
        """
        Close current project and return to a window without an active
        project
        """
        if self.current_active_project:
            path = self.current_active_project.root_path

            self._unmaximize()
            filenames = self.get_plugin()._get_open_filenames()
            if filenames is not None:
                self.set_project_filenames(filenames)

            closed_sucessfully, message = (
                self.current_active_project.close_project())
            if not closed_sucessfully:
                QMessageBox.warning(self, "Project close", message)
                return

            self.current_active_project = None
            self.set_conf('current_project_path', None)
            self._setup_menu_actions()

            with self._disable_pdb_prevent_closing():
                self.sig_project_closed.emit(path)
                self.sig_project_closed[bool].emit(True)

            # Hide pane.
            self.set_conf('visible_if_project_open', self.isVisible())
            self.toggle_view(False)

            self._clear()
            self.sig_restart_console_requested.emit()
            self.watcher.stop()

    def delete_project(self):
        """
        Delete the current project without deleting the files in the directory.
        """
        if self.current_active_project:
            self._unmaximize()
            path = self.current_active_project.root_path

            buttons = QMessageBox.Yes | QMessageBox.No
            answer = QMessageBox.warning(
                self,
                _("Delete"),
                _("Do you really want to delete the <b>{filename}</b> project?"
                  "<br><br>"
                  "<b>Note:</b> This action will only delete the project. "
                  "Its files are going to be preserved on disk."
                  ).format(filename=osp.basename(path)),
                buttons
            )

            if answer == QMessageBox.Yes:
                try:
                    self.close_project()
                    shutil.rmtree(osp.join(path, '.spyproject'))
                except OSError as error:
                    QMessageBox.critical(
                        self,
                        _("Project Explorer"),
                        _("<b>Unable to delete <i>{varpath}</i></b>"
                          "<br><br>The error message was:<br>{error}"
                          ).format(varpath=path, error=str(error))
                    )

                # Remove path from the recent_projects list
                try:
                    self.recent_projects.remove(path)
                except ValueError:
                    pass

    def clear_recent_projects(self):
        """Clear the list of recent projects"""
        self.recent_projects = []
        self.set_conf('recent_projects', self.recent_projects)
        self._setup_menu_actions()

    def change_max_recent_projects(self):
        """Change max recent projects entries."""

        max_projects, valid = QInputDialog.getInt(
            self,
            _('Projects'),
            _('Maximum number of recent projects'),
            self.get_conf('max_recent_projects'),
            1,
            35
        )

        if valid:
            self.set_conf('max_recent_projects', max_projects)

            # This will reduce the number of projects shown in
            # recent_projects_menu according to the new number selected by the
            # user.
            if max_projects < len(self.recent_projects):
                self.recent_projects = self._get_valid_recent_projects(
                    self.recent_projects)[:max_projects]

    def reopen_last_project(self, working_directory, restart_console):
        """
        Reopen the active project when Spyder was closed last time, if any.
        """
        current_project_path = self.get_conf('current_project_path',
                                             default=None)

        # Needs a safer test of project existence!
        if (
            current_project_path and
            self.is_valid_project(current_project_path)
        ):
            self.open_project(
                path=current_project_path,
                restart_console=restart_console,
                save_previous_files=False,
                workdir=working_directory,
            )
            self._load_config()

    def get_project_filenames(self):
        """Get the list of recent filenames of a project"""
        recent_files = []
        if self.current_active_project:
            recent_files = self.current_active_project.get_recent_files()
        elif self.latest_project:
            recent_files = self.latest_project.get_recent_files()
        return recent_files

    def set_project_filenames(self, filenames):
        """Set the list of open file names in a project."""
        if (
            self.current_active_project
            and self.is_valid_project(self.current_active_project.root_path)
        ):
            self.current_active_project.set_recent_files(filenames)

    def is_valid_project(self, path):
        """Check if a directory is a valid Spyder project"""
        spy_project_dir = osp.join(path, '.spyproject')
        return osp.isdir(path) and osp.isdir(spy_project_dir)

    def is_invalid_active_project(self):
        """Handle an invalid active project."""
        try:
            path = self.get_active_project_path()
        except AttributeError:
            return

        if bool(path):
            if not self.is_valid_project(path):
                if path:
                    QMessageBox.critical(
                        self,
                        _('Error'),
                        _("<b>{}</b> is no longer a valid Spyder project! "
                          "Since it is the current active project, it will "
                          "be closed automatically.").format(path)
                    )

                self.close_project()

    def register_project_type(self, parent_plugin, project_type):
        """
        Register a new project type.

        Parameters
        ----------
        parent_plugin: spyder.plugins.api.plugins.SpyderPluginV2
            The parent plugin instance making the project type registration.
        project_type: spyder.plugins.projects.api.BaseProjectType
            Project type to register.
        """
        if not issubclass(project_type, BaseProjectType):
            raise SpyderAPIError("A project type must subclass "
                                 "BaseProjectType!")

        project_id = project_type.ID
        if project_id in self._project_types:
            raise SpyderAPIError("A project type id '{}' has already been "
                                 "registered!".format(project_id))

        project_type._PARENT_PLUGIN = parent_plugin
        self._project_types[project_id] = project_type

    def get_project_types(self):
        """Return available registered project types."""
        return self._project_types

    def get_active_project_path(self):
        """Get path of the active project"""
        if self.current_active_project:
            return self.current_active_project.root_path

    def save_config(self):
        """
        Save configuration: opened projects & tree widget state.

        Also save whether dock widget is visible if a project is open.
        """
        self.set_conf(
            'recent_projects',
            self._get_valid_recent_projects(self.recent_projects)
        )
        self.set_conf('expanded_state', self.treewidget.get_expanded_state())
        self.set_conf('scrollbar_position',
                      self.treewidget.get_scrollbar_position())
        if self.current_active_project:
            self.set_conf('visible_if_project_open', self.isVisible())

    def restore_scrollbar_position(self):
        """Restoring scrollbar position after main window is visible"""
        scrollbar_pos = self.get_conf('scrollbar_position', None)
        if scrollbar_pos is not None:
            self.treewidget.set_scrollbar_position(scrollbar_pos)

    def show_widget(self):
        """Show this widget"""
        self.toggle_view(True)
        self.setVisible(True)
        self.raise_()
        self.update()

    # ---- Public API for the Switcher
    # -------------------------------------------------------------------------
    def display_default_switcher_items(self):
        """Populate switcher with a default set of files in the project."""
        if not self._default_switcher_paths:
            return

        self._display_paths_in_switcher(
            self._default_switcher_paths, setup=False, clear_section=False
        )

    def handle_switcher_selection(self, item, mode, search_text):
        """
        Handle user selecting item in switcher.

        If the selected item is not in the section of the switcher that
        corresponds to this plugin, then ignore it. Otherwise, switch to
        selected project file and hide the switcher.

        Parameters
        ----------
        item: object
            The current selected item from the switcher list (QStandardItem).
        mode: str
            The current selected mode (open files "", symbol "@" or line ":").
        search_text: str
            Cleaned search/filter text.
        """
        if item.get_section() != self.get_title():
            return

        # Open file in editor
        self.sig_open_file_requested.emit(item.get_data())

    def handle_switcher_search(self, search_text):
        """
        Handle user typing in switcher to filter results.

        Load switcher results when a search text is typed for projects.
        Parameters
        ----------
        text: str
            The current search text in the switcher dialog box.
        """
        self._call_fzf(search_text)

    # ---- Public API for the LSP
    # -------------------------------------------------------------------------
    def start_workspace_services(self):
        """Enable LSP workspace functionality."""
        self.completions_available = True
        if self.current_active_project:
            path = self.get_active_project_path()
            self.notify_project_open(path)

    def stop_workspace_services(self):
        """Disable LSP workspace functionality."""
        self.completions_available = False

    def emit_request(self, method, params, requires_response):
        """Send request/notification/response to all LSP servers."""
        params['requires_response'] = requires_response
        params['response_instance'] = self
        self.sig_broadcast_notification_requested.emit(method, params)

    @Slot(str, dict)
    def handle_response(self, method, params):
        """Method dispatcher for LSP requests."""
        if method in self.handler_registry:
            handler_name = self.handler_registry[method]
            handler = getattr(self, handler_name)
            handler(params)

    @request(method=CompletionRequestTypes.WORKSPACE_WATCHED_FILES_UPDATE,
             requires_response=False)
    @Slot(str, bool)
    def file_created(self, src_file, is_dir):
        """Notify LSP server about file creation."""
        self._update_default_switcher_paths()

        # LSP specification only considers file updates
        if is_dir:
            return

        params = {
            'params': [{
                'file': src_file,
                'kind': FileChangeType.CREATED
            }]
        }
        return params

    @Slot(str, str, bool)
    @request(method=CompletionRequestTypes.WORKSPACE_WATCHED_FILES_UPDATE,
             requires_response=False)
    def file_moved(self, src_file, dest_file, is_dir):
        """Notify LSP server about a file that is moved."""
        self._update_default_switcher_paths()

        if is_dir:
            return

        deletion_entry = {
            'file': src_file,
            'kind': FileChangeType.DELETED
        }

        addition_entry = {
            'file': dest_file,
            'kind': FileChangeType.CREATED
        }

        entries = [addition_entry, deletion_entry]
        params = {
            'params': entries
        }
        return params

    @request(method=CompletionRequestTypes.WORKSPACE_WATCHED_FILES_UPDATE,
             requires_response=False)
    @Slot(str, bool)
    def file_deleted(self, src_file, is_dir):
        """Notify LSP server about file deletion."""
        self._update_default_switcher_paths()

        if is_dir:
            return

        params = {
            'params': [{
                'file': src_file,
                'kind': FileChangeType.DELETED
            }]
        }
        return params

    @request(method=CompletionRequestTypes.WORKSPACE_WATCHED_FILES_UPDATE,
             requires_response=False)
    @Slot(str, bool)
    def file_modified(self, src_file, is_dir):
        """Notify LSP server about file modification."""
        if is_dir:
            return

        params = {
            'params': [{
                'file': src_file,
                'kind': FileChangeType.CHANGED
            }]
        }
        return params

    @request(method=CompletionRequestTypes.WORKSPACE_FOLDERS_CHANGE,
             requires_response=False)
    def notify_project_open(self, path):
        """Notify LSP server about project path availability."""
        params = {
            'folder': path,
            'instance': self,
            'kind': 'addition'
        }
        return params

    @request(method=CompletionRequestTypes.WORKSPACE_FOLDERS_CHANGE,
             requires_response=False)
    def notify_project_close(self, path):
        """Notify LSP server to unregister project path."""
        params = {
            'folder': path,
            'instance': self,
            'kind': 'deletion'
        }
        return params

    @handles(CompletionRequestTypes.WORKSPACE_APPLY_EDIT)
    @request(method=CompletionRequestTypes.WORKSPACE_APPLY_EDIT,
             requires_response=False)
    def handle_workspace_edit(self, params):
        """Apply edits to multiple files and notify server about success."""
        edits = params['params']
        response = {
            'applied': False,
            'error': 'Not implemented',
            'language': edits['language']
        }
        return response

    # ---- Private API
    # -------------------------------------------------------------------------
    def _set_project_dir(self, directory):
        """Set the project directory"""
        if directory is not None:
            self.treewidget.set_root_path(osp.dirname(directory))
            self.treewidget.set_folder_names([osp.basename(directory)])
        self.treewidget.setup_project_view()

        index = self.treewidget.get_index(directory)
        if index is not None:
            self.treewidget.setExpanded(self.treewidget.get_index(directory),
                                        True)

    def _clear(self):
        """Show an empty view"""
        self.treewidget.hide()
        self.pane_empty.show()

    def _setup_project(self, directory):
        """Setup project"""
        self.pane_empty.hide()
        self.treewidget.show()

        # Setup the directory shown by the tree
        self._set_project_dir(directory)

    def _unmaximize(self):
        """Unmaximize the currently maximized plugin, if not self."""
        if self.get_plugin().main:
            self.sig_unmaximize_plugin_requested[object].emit(
                self.get_plugin()
            )

    def _load_project_type_class(self, path):
        """
        Load a project type class from the config project folder directly.

        Notes
        -----
        This is done directly, since using the EmptyProject would rewrite the
        value in the constructor. If the project found has not been registered
        as a valid project type, the EmptyProject type will be returned.

        Returns
        -------
        spyder.plugins.projects.api.BaseProjectType
            Loaded project type class.
        """
        fpath = osp.join(
            path, get_project_config_folder(), 'config', WORKSPACE + ".ini")

        project_type_id = EmptyProject.ID
        if osp.isfile(fpath):
            config = configparser.ConfigParser()

            # Catch any possible error when reading the workspace config file.
            # Fixes spyder-ide/spyder#17621
            try:
                config.read(fpath, encoding='utf-8')
            except Exception:
                pass

            # This is necessary to catch an error for projects created in
            # Spyder 4 or older versions.
            # Fixes spyder-ide/spyder#17097
            try:
                project_type_id = config[WORKSPACE].get(
                    "project_type", EmptyProject.ID)
            except Exception:
                pass

        EmptyProject._PARENT_PLUGIN = self.get_plugin()
        project_types = self.get_project_types()
        project_type_class = project_types.get(project_type_id, EmptyProject)
        return project_type_class

    def _add_to_recent(self, project):
        """
        Add an entry to recent projetcs

        We only maintain the list of the 10 most recent projects
        """
        if project not in self.recent_projects:
            self.recent_projects.insert(0, project)

        if len(self.recent_projects) > self.get_conf('max_recent_projects'):
            self.recent_projects.pop(-1)

    def _setup_menu_actions(self):
        """Setup and update the menu actions."""
        self.recent_project_menu.clear_actions()

        if self.recent_projects:
            for project in self.recent_projects:
                if self.is_valid_project(project):
                    if os.name == 'nt':
                        name = project
                    else:
                        name = project.replace(get_home_dir(), '~')

                    try:
                        action = self.get_action(name)
                    except KeyError:
                        action = self.create_action(
                            name,
                            text=name,
                            icon=self.create_icon('project_spyder'),
                            triggered=self._build_opener(project),
                        )

                    self.add_item_to_menu(
                        action,
                        menu=self.recent_project_menu,
                        section=RecentProjectsMenuSections.Recent)

        for item in [
                self.clear_recent_projects_action,
                self.max_recent_action]:
            self.add_item_to_menu(
                item,
                menu=self.recent_project_menu,
                section=RecentProjectsMenuSections.Extras)

        self._update_project_actions()
        self.recent_project_menu.render()

    def _build_opener(self, project):
        """Build function opening passed project"""
        def opener(*args, **kwargs):
            self.open_project(path=project)
        return opener

    def _update_project_actions(self):
        """Update actions of the Projects menu"""
        if self.recent_projects:
            self.clear_recent_projects_action.setEnabled(True)
        else:
            self.clear_recent_projects_action.setEnabled(False)

        active = bool(self.get_active_project_path())
        self.close_project_action.setEnabled(active)
        self.delete_project_action.setEnabled(active)

    def _load_config(self):
        """Load configuration: opened projects & tree widget state"""
        expanded_state = self.get_conf('expanded_state', None)

        # Sometimes the expanded state option may be truncated in .ini file
        # (for an unknown reason), in this case it would be converted to a
        # string by 'userconfig':
        if isinstance(expanded_state, str):
            expanded_state = None
        if expanded_state is not None:
            self.treewidget.set_expanded_state(expanded_state)

    def _get_valid_recent_projects(self, recent_projects):
        """
        Get the list of valid recent projects.

        Parameters
        ----------
        recent_projects: list
            List of recent projects as filesystem paths.
        """
        valid_projects = [
            p for p in recent_projects if self.is_valid_project(p)
        ]

        return valid_projects

    @contextmanager
    def _disable_pdb_prevent_closing(self):
        """
        Context manager to disable the pdb_prevent_closing option before
        opening/closing the previous/current open project files.

        Notes
        -----
        * This is necessary to correctly do that when a console was left in
          debugging mode.
        """
        try:
            pdb_prevent_closing = self.get_conf(
                "pdb_prevent_closing", section="debugger"
            )
            self.set_conf("pdb_prevent_closing", False, section="debugger")
            yield
        finally:
            self.set_conf(
                "pdb_prevent_closing", pdb_prevent_closing, section="debugger"
            )

    # ---- Private API for the Switcher
    # -------------------------------------------------------------------------
    def _call_fzf(self, search_text=""):
        """
        Call fzf in a worker to get the list of files in the current project
        that match with `search_text`.

        Parameters
        ----------
        search_text: str, optional
            The search text to pass to fzf.
        """
        project_path = self.get_active_project_path()
        if (
            not self.get_conf("search_files_in_switcher")
            or self._fzf is None
            or project_path is None
        ):
            return

        self._worker_manager.terminate_all()

        worker = self._worker_manager.create_process_worker(
            [self._fzf, "--filter", search_text],
            os.environ.copy()
        )

        worker.set_cwd(project_path)
        worker.sig_finished.connect(self._process_fzf_output)
        worker.start()

    def _process_fzf_output(self, worker, output, error):
        """Process output that comes from the fzf worker."""
        if output is None or error:
            return

        # Get list of paths from fzf output
        relative_path_list = output.decode('utf-8').strip().split("\n")

        # List of results with absolute path
        if relative_path_list != ['']:
            project_path = self.get_active_project_path()
            result_list = [
                osp.normpath(os.path.join(project_path, path))
                for path in relative_path_list
            ]
        else:
            result_list = []

        # Filter files that can be opened in the editor
        result_list = [
            path for path in result_list
            if osp.splitext(path)[1] in EDIT_EXTENSIONS
        ]

        # Limit the number of results to not introduce lags when displaying
        # them in the switcher.
        if len(result_list) > self.MAX_SWITCHER_RESULTS:
            result_list = result_list[:self.MAX_SWITCHER_RESULTS]

        if not self._default_switcher_paths:
            self._default_switcher_paths = result_list
        else:
            self._display_paths_in_switcher(
                result_list, setup=True, clear_section=True
            )

    def _convert_paths_to_switcher_items(self, paths):
        """
        Convert a list of paths to items that can be shown in the switcher.
        """
        # The paths that are opened in the editor need to be excluded because
        # they are already shown in the Editor section of the switcher.
        open_files = self.get_plugin()._get_open_filenames()
        for file in open_files:
            normalized_path = osp.normpath(file)
            if normalized_path in paths:
                paths.remove(normalized_path)

        is_unsaved = [False] * len(paths)
        short_paths = shorten_paths(paths, is_unsaved)
        section = self.get_title()

        items = []
        for i, (path, short_path) in enumerate(zip(paths, short_paths)):
            title = osp.basename(path)
            icon = get_file_icon(path)
            description = osp.dirname(path)
            if len(path) > 75:
                description = short_path
            is_last_item = (i + 1 == len(paths))

            item_tuple = (
                title, description, icon, section, path, is_last_item
            )
            items.append(item_tuple)

        return items

    def _display_paths_in_switcher(self, paths, setup, clear_section):
        """Display a list of paths in the switcher."""
        items = self._convert_paths_to_switcher_items(paths)

        # Call directly the plugin's method instead of emitting a signal
        # because it's faster.
        self._plugin._display_items_in_switcher(items, setup, clear_section)

    def _clear_switcher_paths(self):
        """Clear saved switcher results."""
        self._default_switcher_paths = []

    def _update_default_switcher_paths(self):
        """Update default paths to be shown in the switcher."""
        self._default_switcher_paths = []
        self._call_fzf()

    @on_conf_change(option="search_files_in_switcher")
    def _on_search_files_in_switcher_changed(self, value):
        """
        Actions to take when users enable/disable searching files in the
        switcher.
        """
        if value:
            self._update_default_switcher_paths()
        else:
            self._clear_switcher_paths()


# =============================================================================
# Tests
# =============================================================================
class ProjectExplorerTest(QWidget):
    def __init__(self, directory=None):
        QWidget.__init__(self)
        self.CONF_SECTION = 'project_explorer'
        vlayout = QVBoxLayout()
        self.setLayout(vlayout)

        self.explorer = ProjectExplorerWidget(None, self, self)
        if directory is not None:
            self.directory = directory
        else:
            self.directory = osp.dirname(osp.abspath(__file__))
        self.explorer._setup_project(self.directory)
        vlayout.addWidget(self.explorer)

        hlayout1 = QHBoxLayout()
        vlayout.addLayout(hlayout1)
        label = QLabel("<b>Open file:</b>")
        label.setAlignment(Qt.AlignRight)
        hlayout1.addWidget(label)
        self.label1 = QLabel()
        hlayout1.addWidget(self.label1)
        self.explorer.sig_open_file_requested.connect(self.label1.setText)

        hlayout3 = QHBoxLayout()
        vlayout.addLayout(hlayout3)
        label = QLabel("<b>Option changed:</b>")
        label.setAlignment(Qt.AlignRight)
        hlayout3.addWidget(label)
        self.label3 = QLabel()
        hlayout3.addWidget(self.label3)


def test():
    from spyder.utils.qthelpers import qapplication
    app = qapplication()
    test = ProjectExplorerTest()
    test.resize(250, 480)
    test.show()
    app.exec_()


if __name__ == "__main__":
    test()



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder___init__.py for details)

"""IPython console run executor configuration widget."""

# Standard library imports
import os.path as osp

# Third-party imports
from qtpy.compat import getexistingdirectory
from qtpy.QtWidgets import (
    QCheckBox, QGroupBox, QHBoxLayout, QLineEdit, QRadioButton, QVBoxLayout)

# Local imports
from spyder.api.translations import _
from spyder.plugins.run.api import (
    RunExecutorConfigurationGroup, Context, RunConfigurationMetadata)
from spyder.utils.misc import getcwd_or_home


# Main constants
CURRENT_INTERPRETER = _("Execute in current console")
DEDICATED_INTERPRETER = _("Execute in a dedicated console")
CLEAR_ALL_VARIABLES = _("Remove all variables before execution")
CONSOLE_NAMESPACE = _("Run in console's namespace instead of an empty one")
POST_MORTEM = _("Directly enter debugging when errors appear")


class IPythonConfigOptions(RunExecutorConfigurationGroup):
    """IPython console run configuration options."""

    def __init__(self, parent, context: Context, input_extension: str,
                 input_metadata: RunConfigurationMetadata):
        super().__init__(parent, context, input_extension, input_metadata)

        self.dir = None

        # --- Interpreter ---
        interpreter_group = QGroupBox(_("Console"))
        interpreter_layout = QVBoxLayout(interpreter_group)

        self.current_radio = QRadioButton(CURRENT_INTERPRETER)
        interpreter_layout.addWidget(self.current_radio)

        self.dedicated_radio = QRadioButton(DEDICATED_INTERPRETER)
        interpreter_layout.addWidget(self.dedicated_radio)

        # --- General settings ----
        common_group = QGroupBox(_("Advanced settings"))
        common_layout = QVBoxLayout(common_group)

        self.clear_var_cb = QCheckBox(CLEAR_ALL_VARIABLES)
        common_layout.addWidget(self.clear_var_cb)

        self.console_ns_cb = QCheckBox(CONSOLE_NAMESPACE)
        common_layout.addWidget(self.console_ns_cb)

        self.post_mortem_cb = QCheckBox(POST_MORTEM)
        common_layout.addWidget(self.post_mortem_cb)

        self.clo_cb = QCheckBox(_("Command line options:"))
        self.clo_edit = QLineEdit(self)
        self.clo_edit.setMinimumWidth(300)
        self.clo_cb.toggled.connect(self.clo_edit.setEnabled)
        self.clo_edit.setEnabled(False)

        cli_layout = QHBoxLayout()
        cli_layout.addWidget(self.clo_cb)
        cli_layout.addWidget(self.clo_edit)
        common_layout.addLayout(cli_layout)

        layout = QVBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(interpreter_group)
        layout.addWidget(common_group)
        layout.addStretch(100)

    def select_directory(self):
        """Select directory"""
        basedir = str(self.wd_edit.text())
        if not osp.isdir(basedir):
            basedir = getcwd_or_home()
        directory = getexistingdirectory(self, _("Select directory"), basedir)
        if directory:
            self.wd_edit.setText(directory)
            self.dir = directory

    @staticmethod
    def get_default_configuration() -> dict:
        return {
            'current': True,
            'post_mortem': False,
            'python_args_enabled': False,
            'python_args': '',
            'clear_namespace': False,
            'console_namespace': False,
        }

    def set_configuration(self, config: dict):
        use_current_console = config['current']
        post_mortem = config['post_mortem']
        py_args_enabled = config['python_args_enabled']
        py_args = config['python_args']
        clear_namespace = config['clear_namespace']
        console_namespace = config['console_namespace']

        self.current_radio.setChecked(use_current_console)
        self.dedicated_radio.setChecked(not use_current_console)
        self.post_mortem_cb.setChecked(post_mortem)
        self.clo_cb.setChecked(py_args_enabled)
        self.clo_edit.setText(py_args)
        self.clear_var_cb.setChecked(clear_namespace)
        self.console_ns_cb.setChecked(console_namespace)

    def get_configuration(self) -> dict:
        return {
            'current': self.current_radio.isChecked(),
            'post_mortem': self.post_mortem_cb.isChecked(),
            'python_args_enabled': self.clo_cb.isChecked(),
            'python_args': self.clo_edit.text(),
            'clear_namespace': self.clear_var_cb.isChecked(),
            'console_namespace': self.console_ns_cb.isChecked(),
        }



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""
Spyder API auxiliary widgets.
"""

# Third party imports
from qtpy.QtCore import QEvent, QSize, Signal
from qtpy.QtWidgets import QMainWindow, QSizePolicy, QToolBar, QWidget

# Local imports
from spyder.api.exceptions import SpyderAPIError
from spyder.api.widgets import PluginMainWidgetWidgets
from spyder.api.widgets.mixins import SpyderMainWindowMixin
from spyder.utils.stylesheet import APP_STYLESHEET


class SpyderWindowWidget(QMainWindow, SpyderMainWindowMixin):
    """MainWindow subclass that contains a SpyderDockablePlugin."""

    # ---- Signals
    # ------------------------------------------------------------------------
    sig_closed = Signal()
    """This signal is emitted when the close event is fired."""

    sig_window_state_changed = Signal(object)
    """
    This signal is emitted when the window state has changed (for instance,
    between maximized and minimized states).

    Parameters
    ----------
    window_state: Qt.WindowStates
        The window state.
    """

    def __init__(self, widget):
        super().__init__()
        self.widget = widget

        # To distinguish these windows from the main Spyder one
        self.is_window_widget = True

        # Setting interface theme
        self.setStyleSheet(str(APP_STYLESHEET))

    def closeEvent(self, event):
        """Override Qt method to emit a custom `sig_close` signal."""
        super().closeEvent(event)
        self.sig_closed.emit()

    def changeEvent(self, event):
        """
        Override Qt method to emit a custom `sig_windowstate_changed` signal
        when there's a change in the window state.
        """
        if event.type() == QEvent.WindowStateChange:
            self.sig_window_state_changed.emit(self.windowState())
        super().changeEvent(event)


class MainCornerWidget(QToolBar):
    """
    Corner widget to hold options menu, spinner and additional options.
    """

    def __init__(self, parent, name):
        super().__init__(parent)
        self._icon_size = QSize(16, 16)
        self.setIconSize(self._icon_size)

        self._widgets = {}
        self._actions = []
        self.setObjectName(name)

        # We add an strut widget here so that there is a spacing
        # between the first item of the corner widget and the last
        # item of the MainWidgetToolbar.
        self._strut = QWidget()
        self._strut.setFixedWidth(0)
        self._strut.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        self.addWidget(self._strut)

    def add_widget(self, widget, before=None):
        """
        Add a widget to the left of the last widget added to the corner.
        """
        if not hasattr(widget, "name") or (
            before is not None and not hasattr(before, "name")
        ):
            raise SpyderAPIError(
                f"Widget {widget} or {before} doesn't have a name, which must "
                f"be provided by the attribute `name`"
            )

        if widget.name in self._widgets:
            raise SpyderAPIError(
                'Wigdet with name "{}" already added. Current names are: {}'
                ''.format(widget.name, list(self._widgets.keys()))
            )

        if before is not None and before.name not in self._widgets:
            raise SpyderAPIError(
                f"Wigdet with name '{before.name}' not in this corner widget"
            )

        if (
            not self._widgets
            and widget.name != PluginMainWidgetWidgets.OptionsToolButton
        ):
            raise SpyderAPIError(
                "The options button must be the first one to be added to the "
                "corner widget of dockable plugins."
            )

        if widget.name == PluginMainWidgetWidgets.OptionsToolButton:
            # This is only necessary for the options button because it's the
            # first one to be added
            action = self.addWidget(widget)
        else:
            if before is not None:
                before_action = self.get_action(before.name)
            else:
                # By default other buttons are added to the left of the last
                # one
                before_action = self._actions[-1]

            # Allow to add either widgets or actions
            if isinstance(widget, QWidget):
                action = self.insertWidget(before_action, widget)
            else:
                action = widget
                self.insertAction(before_action, action)
                widget = self.widgetForAction(action)
                widget.name = action.name

        self._widgets[widget.name] = (widget, action)
        self._actions.append(action)

    def get_widget(self, widget_id):
        """Return a widget by unique id."""
        if widget_id in self._widgets:
            return self._widgets[widget_id][0]

    def get_action(self, widget_id):
        """Return action corresponding to `widget_id`."""
        if widget_id in self._widgets:
            return self._widgets[widget_id][1]



from math import cos
from numpy import (
   linspace)

from subdir.a import (
    bar)

# We shouldn't show symbols for relative imports in the Outline.
# This is a regression test for issue spyder-ide/spyder#16352.
from ..a import (
    MyOtherClass)

from ...file2 import (
    MyClass,
    foo
)

def baz(x):
    return x

class AnotherClass:
    E = 1

    def five(self):
        return 5

    def six(self):
        return 4



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
#

"""Tests for the Array Builder Widget."""

# Standard library imports
import sys

# Third party imports
from qtpy.QtCore import Qt
import pytest

# Local imports
from spyder.widgets.arraybuilder import ArrayBuilderDialog


# --- Fixtures
# -----------------------------------------------------------------------------
@pytest.fixture
def botinline(qtbot):
    dialog = ArrayBuilderDialog(inline=True)
    qtbot.addWidget(dialog)
    dialog.show()
    return dialog, dialog.array_widget


@pytest.fixture
def botinlinefloat(qtbot):
    dialog = ArrayBuilderDialog(inline=True, force_float=True)
    qtbot.addWidget(dialog)
    dialog.show()
    return dialog, dialog.array_widget


@pytest.fixture
def botarray(qtbot):
    dialog = ArrayBuilderDialog(inline=False)
    qtbot.addWidget(dialog)
    dialog.show()
    return dialog, dialog.array_widget


# --- Tests
# -----------------------------------------------------------------------------
def test_array_inline(botinline, qtbot):
    dialog, widget = botinline
    qtbot.keyClicks(widget, '1 2 3; 4 5 6')
    qtbot.keyPress(widget, Qt.Key_Return, modifier=Qt.ShiftModifier)
    value = dialog.text()
    assert value == 'np.array([[1, 2, 3],\n          [4, 5, 6]])'


def test_array_inline_array_invalid(botinline, qtbot):
    dialog, widget = botinline
    qtbot.keyClicks(widget, '1 2; 3 4; 5 6 7')
    qtbot.keyPress(widget, Qt.Key_Return, modifier=Qt.ShiftModifier)
    assert not dialog.is_valid()


def test_array_inline_1d_array(botinline, qtbot):
    dialog, widget = botinline
    qtbot.keyClicks(widget, '4 5 6')
    qtbot.keyPress(widget, Qt.Key_Return, modifier=Qt.ShiftModifier)
    value = dialog.text()
    assert value == 'np.array([4, 5, 6])'


def test_array_inline_nan_array(botinline, qtbot):
    dialog, widget = botinline
    qtbot.keyClicks(widget, '4 nan 6 8 9')
    qtbot.keyPress(widget, Qt.Key_Return, modifier=Qt.ShiftModifier)
    value = dialog.text()
    assert value == 'np.array([4, np.nan, 6, 8, 9])'


def test_array_inline_inf_array(botinline, qtbot):
    dialog, widget = botinline
    qtbot.keyClicks(widget, '4 inf 6 8 9')
    qtbot.keyPress(widget, Qt.Key_Return, modifier=Qt.ShiftModifier)
    value = dialog.text()
    assert value == 'np.array([4, np.inf, 6, 8, 9])'


def test_array_inline_force_float_array(botinlinefloat, qtbot):
    dialog, widget = botinlinefloat
    qtbot.keyClicks(widget, '4 5 6 8 9')
    qtbot.keyPress(widget, Qt.Key_Return, modifier=Qt.ShiftModifier)
    value = dialog.text()
    assert value == 'np.array([4.0, 5.0, 6.0, 8.0, 9.0])'


def test_array_inline_force_float_error_array(botinlinefloat, qtbot):
    dialog, widget = botinlinefloat
    qtbot.keyClicks(widget, '4 5 6 a 9')
    qtbot.keyPress(widget, Qt.Key_Return, modifier=Qt.ShiftModifier)
    value = dialog.text()
    assert value == 'np.array([4.0, 5.0, 6.0, a, 9.0])'


@pytest.mark.skipif(sys.platform == 'darwin', reason="Fails on macOS")
def test_array_table(botarray, qtbot):
    dialog, widget = botarray
    qtbot.keyClick(widget, Qt.Key_1)
    qtbot.keyClick(widget, Qt.Key_Tab)
    qtbot.keyClick(widget, Qt.Key_2)
    qtbot.keyClick(widget, Qt.Key_Tab)
    qtbot.keyClick(widget, Qt.Key_Backtab)  # Hack: in the tests the selected cell is wrong
    qtbot.keyClick(widget, Qt.Key_3)
    qtbot.keyClick(widget, Qt.Key_Tab)
    qtbot.keyClick(widget, Qt.Key_4)
    qtbot.keyClick(widget, Qt.Key_Tab)
    qtbot.keyClick(widget, Qt.Key_5)
    qtbot.keyClick(widget, Qt.Key_Tab)
    qtbot.keyClick(widget, Qt.Key_6)
    qtbot.keyClick(widget, Qt.Key_Tab)  # Hack: in the tests the selected cell is wrong
    qtbot.keyClick(widget, Qt.Key_Return, modifier=Qt.ShiftModifier)
    value = dialog.text()
    assert value == 'np.array([[1, 2, 3],\n          [4, 5, 6]])'


@pytest.mark.skipif(sys.platform == 'darwin', reason="Fails on macOS")
def test_array_table_empty_items(botarray, qtbot):  # analysis:ignore
    dialog, widget = botarray
    qtbot.keyClick(widget, Qt.Key_Tab)
    qtbot.keyClick(widget, Qt.Key_2)
    qtbot.keyClick(widget, Qt.Key_Tab)
    qtbot.keyClick(widget, Qt.Key_Backtab)  # Hack: in the tests the selected cell is wrong
    qtbot.keyClick(widget, Qt.Key_3)
    qtbot.keyClick(widget, Qt.Key_Tab)
    qtbot.keyClick(widget, Qt.Key_Tab)
    qtbot.keyClick(widget, Qt.Key_5)
    qtbot.keyClick(widget, Qt.Key_Tab)
    qtbot.keyClick(widget, Qt.Key_6)
    qtbot.keyClick(widget, Qt.Key_Tab)  # Hack: in the tests the selected cell is wrong
    qtbot.keyClick(widget, Qt.Key_Return, modifier=Qt.ShiftModifier)
    value = dialog.text()
    assert value == 'np.array([[0, 2, 3],\n          [0, 5, 6]])'


@pytest.mark.skipif(sys.platform == 'darwin', reason="Fails on macOS")
def test_array_table_spaces_in_item(botarray, qtbot):  # analysis:ignore
    dialog, widget = botarray
    qtbot.keyClicks(widget, '   ')
    qtbot.keyClick(widget, Qt.Key_Tab)
    qtbot.keyClick(widget, Qt.Key_2)
    qtbot.keyClick(widget, Qt.Key_Tab)
    qtbot.keyClick(widget, Qt.Key_Backtab)
    qtbot.keyClick(widget, Qt.Key_3)
    qtbot.keyClick(widget, Qt.Key_Tab)
    qtbot.keyClick(widget, Qt.Key_Tab)
    qtbot.keyClick(widget, Qt.Key_5)
    qtbot.keyClick(widget, Qt.Key_Tab)
    qtbot.keyClick(widget, Qt.Key_6)
    qtbot.keyClick(widget, Qt.Key_Tab)  # Hack: in the tests the selected cell is wrong
    qtbot.keyClick(widget, Qt.Key_Return, modifier=Qt.ShiftModifier)
    value = dialog.text()
    assert value == 'np.array([[0, 2, 3],\n          [0, 5, 6]])'


@pytest.mark.skipif(sys.platform == 'darwin', reason="Fails on macOS")
def test_array_table_empty(botarray, qtbot):  # analysis:ignore
    dialog, widget = botarray
    qtbot.keyClick(widget, Qt.Key_Return, modifier=Qt.ShiftModifier)
    value = dialog.text()
    assert value == ''



# -*- coding: utf-8 -*-

# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

from unittest.mock import Mock, MagicMock

from qtpy.QtCore import QObject, Signal, Slot
from qtpy.QtWidgets import QWidget
import pytest
from pytestqt.qtbot import QtBot

from spyder.config.manager import CONF
from spyder.plugins.completion.api import SERVER_CAPABILITES
from spyder.plugins.completion.tests.conftest import qtbot_module
from spyder.plugins.completion.providers.languageserver.provider import (
    LanguageServerProvider)


class CompletionPluginMock(QObject, MagicMock):
    """Mock for the completion plugin."""
    CONF_SECTION = 'completions'

    def __init__(self, conf):
        super().__init__()
        self.conf = conf

    def get_conf(self, option, default=None, section=None):
        if section == 'completions':
            option = option[-1]
            return self.conf.get(option, default)
        else:
            return CONF.get(section, option, default)


def lsp_context(is_stdio):
    @pytest.fixture(scope='module')
    def wrapper(qtbot_module, request):
        # Activate pycodestyle and pydocstyle
        conf = dict(LanguageServerProvider.CONF_DEFAULTS)
        conf['pycodestyle'] = True
        conf['pydocstyle'] = True
        conf['stdio'] = is_stdio

        # Create the manager
        provider = LanguageServerProvider(CompletionPluginMock(conf), conf)

        # Wait for the client to be started
        with qtbot_module.waitSignal(
                provider.sig_language_completions_available,
                timeout=30000) as block:
            provider.start_completion_services_for_language('python')

        capabilities, _ = block.args
        assert all(
            [option in SERVER_CAPABILITES for option in capabilities.keys()])

        def teardown():
            provider.shutdown()

        request.addfinalizer(teardown)
        return provider
    return wrapper


lsp_provider = lsp_context(is_stdio=False)
lsp_stdio_provider = lsp_context(is_stdio=True)



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""Run configuration page."""

# Standard library imports
import functools
from copy import deepcopy
from typing import Dict, List, Set, Tuple
from uuid import uuid4

# Third party imports
from qtpy.QtCore import Qt
from qtpy.QtWidgets import (
    QAbstractItemView,
    QGroupBox,
    QHBoxLayout,
    QHeaderView,
    QLabel,
    QVBoxLayout,
)

# Local imports
from spyder.api.preferences import PluginConfigPage
from spyder.api.translations import _
from spyder.api.widgets.comboboxes import SpyderComboBox
from spyder.plugins.run.api import (
    ExtendedRunExecutionParameters, SupportedExecutionRunConfiguration)
from spyder.plugins.run.container import RunContainer
from spyder.plugins.run.models import (
    RunExecutorNamesListModel, ExecutorRunParametersTableModel)
from spyder.plugins.run.widgets import (
    ExecutionParametersDialog, RunDialogStatus)
from spyder.utils.icon_manager import ima
from spyder.utils.stylesheet import AppStyle
from spyder.widgets.helperwidgets import HoverRowsTableView


def move_file_to_front(contexts: List[str]) -> List[str]:
    if 'file' in contexts:
        contexts.insert(0, contexts.pop(contexts.index('file')))
    return contexts


class RunParametersTableView(HoverRowsTableView):

    def __init__(self, parent, model):
        super().__init__(parent)
        self._parent = parent
        self.setModel(model)
        self.setSelectionBehavior(QAbstractItemView.SelectRows)
        self.setSelectionMode(QAbstractItemView.SingleSelection)
        self.setSortingEnabled(True)
        self.setEditTriggers(QAbstractItemView.AllEditTriggers)
        self.selectionModel().selectionChanged.connect(self.selection)
        self.verticalHeader().hide()
        self.reset_plain()

        self.horizontalHeader().setSectionResizeMode(0, QHeaderView.Stretch)
        self.horizontalHeader().setSectionResizeMode(
            1, QHeaderView.Stretch
        )
        self.horizontalHeader().setSectionResizeMode(2, QHeaderView.Stretch)

    def selection(self, index):
        self.update()
        self.isActiveWindow()

        # Detect if a row corresponds to a set of default parameters to prevent
        # users from deleting it.
        index = self.currentIndex().row()
        is_default = False
        if index >= 0:
            params_id = self._parent.table_model.params_index[index]
            params = self._parent.table_model.executor_conf_params[params_id]
            is_default = True if params.get("default") else False

        self._parent.set_buttons_status(is_default=is_default)

        # Always enable edit button
        self._parent.edit_configuration_btn.setEnabled(True)

    def adjust_cells(self):
        """Adjust column size based on contents."""
        self.resizeColumnsToContents()
        model: ExecutorRunParametersTableModel = self.model()
        fm = self.horizontalHeader().fontMetrics()
        names = [fm.width(model.get_parameters_name(idx)) for idx in model]
        if names:
            self.setColumnWidth(
                ExecutorRunParametersTableModel.NAME, max(names))
        self.horizontalHeader().setStretchLastSection(True)

    def reset_plain(self):
        self.model().reset_model()
        self.adjust_cells()
        self.selectionModel().selectionChanged.connect(self.selection)

    def show_editor(self, new=False, clone=False):
        extension, context, params = None, None, None
        extensions, contexts, plugin_name, executor_params = (
            self._parent.get_executor_configurations()
        )

        if not new:
            index = self.currentIndex().row()
            model: ExecutorRunParametersTableModel = self.model()
            (extension, context, params) = model[index]

        self.dialog = ExecutionParametersDialog(
            parent=self,
            executor_name=plugin_name,
            executor_params=executor_params,
            param_names=self.model().get_parameter_names(),
            extensions=extensions,
            contexts=contexts,
            current_params=params,
            extension=extension,
            context=context,
            new_config=new,
        )

        self.dialog.setup()
        self.dialog.finished.connect(
            functools.partial(
                self.process_run_dialog_result,
                new=new, clone=clone, params=params
            )
        )

        if not clone:
            self.dialog.open()
        else:
            self.dialog.accept()

    def process_run_dialog_result(self, result, new=False,
                                  clone=False, params=None):
        status = self.dialog.status
        if status == RunDialogStatus.Close:
            return

        conf = self.dialog.get_configuration()
        if conf is None:
            return
        else:
            extension, context, new_executor_params = conf

        if not new and clone:
            new_executor_params["uuid"] = str(uuid4())
            new_executor_params["name"] = _('%s (copy)') % params['name']

        changes = []
        if params and not clone:
            changes.append(('deleted', params))

        changes.append(('new', new_executor_params))
        self.model().apply_changes(changes, extension, context)

    def clone_configuration(self):
        self.show_editor(clone=True)

    def focusInEvent(self, e):
        """Qt Override."""
        super().focusInEvent(e)
        self.selectRow(self.currentIndex().row())
        self.selection(self.currentIndex().row())

    def keyPressEvent(self, event):
        """Qt Override."""
        key = event.key()
        if key in [Qt.Key_Enter, Qt.Key_Return]:
            self.show_editor()
        elif key in [Qt.Key_Up, Qt.Key_Down, Qt.Key_Left, Qt.Key_Right]:
            super().keyPressEvent(event)
        else:
            super().keyPressEvent(event)

    def mouseDoubleClickEvent(self, event):
        """Qt Override."""
        self.show_editor()


class RunConfigPage(PluginConfigPage):
    """Default Run Settings configuration page."""

    def setup_page(self):
        self._params_to_delete = {}

        # --- Executors tab ---
        self.plugin_container: RunContainer = self.plugin.get_container()
        self.executor_model = RunExecutorNamesListModel(
            self, self.plugin_container.executor_model)
        self.table_model = ExecutorRunParametersTableModel(self)
        self.table_model.sig_data_changed.connect(
            self.on_table_data_changed
        )

        self.all_executor_model: Dict[
            str, Dict[Tuple[str, str, str],
            ExtendedRunExecutionParameters]] = {}
        self.previous_executor_index: int = 0
        self.default_executor_conf_params: Dict[
            str, Dict[Tuple[str, str, str],
            ExtendedRunExecutionParameters]] = {}

        about_label = QLabel(
            _(
                "The following are the global configuration presets of the "
                "different runners that can execute files in Spyder."
            )
        )
        about_label.setWordWrap(True)

        # The paremeters table needs to be created before the executor_combo
        # below, although is displayed after it.
        params_label = QLabel(_('Configuration presets:'))
        self.params_table = RunParametersTableView(self, self.table_model)
        self.params_table.setMaximumHeight(180)

        params_table_layout = QHBoxLayout()
        params_table_layout.addSpacing(2 * AppStyle.MarginSize)
        params_table_layout.addWidget(self.params_table)
        params_table_layout.addSpacing(2 * AppStyle.MarginSize)

        executor_label = QLabel(_("Runner:"))
        self.executor_combo = SpyderComboBox(self)
        self.executor_combo.setMinimumWidth(250)
        self.executor_combo.currentIndexChanged.connect(
            self.executor_index_changed
        )
        self.executor_combo.setModel(self.executor_model)

        executor_layout = QHBoxLayout()
        executor_layout.addWidget(executor_label)
        executor_layout.addWidget(self.executor_combo)
        executor_layout.addStretch()

        # Buttons
        self.new_configuration_btn = self.create_button(
            icon=ima.icon("edit_add"),
            callback=self.create_new_configuration,
            tooltip=_("New parameters"),
        )
        self.edit_configuration_btn = self.create_button(
            icon=ima.icon("edit"),
            callback=self.edit_configuration,
            tooltip=_("Edit selected"),
        )
        self.clone_configuration_btn = self.create_button(
            icon=ima.icon("editcopy"),
            callback=self.clone_configuration,
            tooltip=_("Clone selected"),
        )
        self.delete_configuration_btn = self.create_button(
            icon=ima.icon("editclear"),
            callback=self.delete_configuration,
            tooltip=_("Delete selected"),
        )
        self.reset_changes_btn = self.create_button(
            icon=ima.icon("restart"),
            callback=self.reset_changes,
            tooltip=_("Reset current changes"),
        )

        # Disable edition button at startup
        self.set_buttons_status(status=False)

        # Buttons layout
        buttons_layout = QHBoxLayout()
        buttons_layout.addStretch()

        btns = [
            self.new_configuration_btn,
            self.edit_configuration_btn,
            self.delete_configuration_btn,
            self.clone_configuration_btn,
            self.reset_changes_btn,
        ]
        for btn in btns:
            buttons_layout.addWidget(btn)

        buttons_layout.addStretch()

        # Final layout
        presets_group = QGroupBox(_("Global presets"))
        presets_layout = QVBoxLayout()
        presets_group.setLayout(presets_layout)

        presets_layout.addWidget(about_label)
        presets_layout.addSpacing(3 * AppStyle.MarginSize)
        presets_layout.addLayout(executor_layout)
        presets_layout.addSpacing(3 * AppStyle.MarginSize)
        presets_layout.addWidget(params_label)
        presets_layout.addLayout(params_table_layout)
        presets_layout.addSpacing(AppStyle.MarginSize)
        presets_layout.addLayout(buttons_layout)

        # --- Editor interactions tab ---
        newcb = self.create_checkbox
        saveall_box = newcb(_("Save all files before running script"),
                            'save_all_before_run')
        run_cell_box = newcb(_("Copy full cell contents to the console when "
                               "running code cells"), 'run_cell_copy')

        run_group = QGroupBox(_("Editor interactions"))
        run_layout = QVBoxLayout()
        run_group.setLayout(run_layout)

        run_layout.addWidget(saveall_box)
        run_layout.addWidget(run_cell_box)

        # --- Page layout ----
        vlayout = QVBoxLayout()
        vlayout.addWidget(presets_group)
        vlayout.addWidget(run_group)
        vlayout.addStretch()
        self.setLayout(vlayout)

    def executor_index_changed(self, index: int):
        # Save previous executor configuration
        prev_executor_info = self.table_model.get_current_view()
        previous_executor_name, _ = self.executor_model.selected_executor(
            self.previous_executor_index)
        self.all_executor_model[previous_executor_name] = prev_executor_info

        # Handle current executor configuration
        executor, available_inputs = self.executor_model.selected_executor(
            index)
        container = self.plugin_container

        executor_conf_params = self.all_executor_model.get(executor, {})
        if executor_conf_params == {}:
            for (ext, context) in available_inputs:
                params = container.get_executor_configuration_parameters(
                    executor, ext, context
                )
                params = params["params"]
                for exec_params_id in params:
                    exec_params = params[exec_params_id]

                    # Don't display configs set for specific files. Here
                    # users are allowed to configure global configs, i.e. those
                    # that can be used by any file.
                    if exec_params.get("file_uuid") is not None:
                        continue

                    params_key = (ext, context, exec_params_id)
                    executor_conf_params[params_key] = exec_params

            self.default_executor_conf_params[executor] = deepcopy(
                executor_conf_params)

            self.all_executor_model[executor] = deepcopy(executor_conf_params)

        self.table_model.set_parameters(executor_conf_params)
        self.previous_executor_index = index
        self.set_buttons_status()

    def on_table_data_changed(self):
        # Buttons need to be disabled because the table model is reset when
        # data is changed and focus is lost
        self.set_buttons_status(False)
        self.set_modified(True)

    def set_buttons_status(self, status=None, is_default=False):
        # We need to enclose the code below in a try/except because these
        # buttons might not be created yet, which gives an AttributeError.
        try:
            if status is None:
                status = (
                    self.table_model.rowCount() != 0
                    and self.params_table.currentIndex().isValid()
                )

            # Don't allow to delete default configurations
            if is_default:
                self.delete_configuration_btn.setEnabled(False)
            else:
                self.delete_configuration_btn.setEnabled(status)

            self.edit_configuration_btn.setEnabled(status)
            self.clone_configuration_btn.setEnabled(status)
        except AttributeError:
            pass

    def get_executor_configurations(self) -> Dict[
            str, SupportedExecutionRunConfiguration]:
        exec_index = self.executor_combo.currentIndex()
        executor_name, available_inputs = (
            self.executor_model.selected_executor(exec_index)
        )

        executor_params: Dict[str, SupportedExecutionRunConfiguration] = {}
        extensions: Set[str] = set({})
        contexts: Dict[str, List[str]] = {}

        conf_indices = (
            self.plugin_container.executor_model.executor_configurations
        )

        for _input in available_inputs:
            extension, context = _input
            extensions |= {extension}
            ext_contexts = contexts.get(extension, [])
            ext_contexts.append(context)
            contexts[extension] = ext_contexts

            executors = conf_indices[_input]
            conf = executors[executor_name]
            executor_params[_input] = conf

        contexts = {
            ext: move_file_to_front(ctx) for ext, ctx in contexts.items()
        }

        # Localized version of the executor
        executor_loc_name = self.main.get_plugin(executor_name).get_name()

        return (
            list(sorted(extensions)),
            contexts,
            executor_loc_name,
            executor_params
        )

    def create_new_configuration(self):
        self.params_table.show_editor(new=True)

    def edit_configuration(self):
        self.params_table.show_editor()

    def clone_configuration(self):
        self.params_table.clone_configuration()

    def delete_configuration(self):
        executor_name, __ = self.executor_model.selected_executor(
            self.previous_executor_index
        )
        index = self.params_table.currentIndex().row()
        conf_index = self.table_model.get_tuple_index(index)

        executor_params = self.table_model.executor_conf_params
        executor_params.pop(conf_index, None)

        if executor_name not in self._params_to_delete:
            self._params_to_delete[executor_name] = []
        self._params_to_delete[executor_name].append(conf_index)

        self.table_model.set_parameters(executor_params)
        self.table_model.reset_model()

        self.set_modified(True)
        self.set_buttons_status()

    def reset_changes(self):
        """Reset changes to the parameters loaded when the page was created."""
        self.all_executor_model = deepcopy(self.default_executor_conf_params)
        executor_name, _ = self.executor_model.selected_executor(
            self.previous_executor_index
        )
        executor_params = self.all_executor_model[executor_name]
        self.table_model.set_parameters(executor_params)
        self.table_model.reset_model()
        self.set_modified(True)
        self.set_buttons_status()

    def apply_settings(self):
        prev_executor_info = self.table_model.get_current_view()
        previous_executor_name, _ = self.executor_model.selected_executor(
            self.previous_executor_index
        )
        self.all_executor_model[previous_executor_name] = prev_executor_info

        # Save new parameters
        for executor in self.all_executor_model:
            executor_params = self.all_executor_model[executor]
            stored_execution_params: Dict[
                Tuple[str, str],
                Dict[str, ExtendedRunExecutionParameters]] = {}

            for key in executor_params:
                (extension, context, params_id) = key
                params = executor_params[key]
                ext_ctx_list = stored_execution_params.get(
                    (extension, context), {})
                ext_ctx_list[params_id] = params
                stored_execution_params[(extension, context)] = ext_ctx_list

            for extension, context in stored_execution_params:
                ext_ctx_list = stored_execution_params[(extension, context)]
                self.plugin_container.set_executor_configuration_parameters(
                    executor, extension, context, {'params': ext_ctx_list}
                )

        # Delete removed parameters
        for executor in self._params_to_delete:
            executor_params_to_delete = self._params_to_delete[executor]

            for key in executor_params_to_delete:
                (extension, context, params_id) = key
                self.plugin_container.delete_executor_configuration_parameters(
                    executor, extension, context, params_id
                )

        self._params_to_delete = {}

        # This is necessary to prevent giving focus to the executor combobox,
        # which is odd.
        self.setFocus()

        return {'parameters'}



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
#

"""
Tests for the Spyder kernel
"""

import os
import pytest

from spyder.config.manager import CONF
from spyder.plugins.ipythonconsole.utils.kernelspec import SpyderKernelSpec
from spyder.py3compat import to_text_string


@pytest.mark.parametrize('default_interpreter', [True, False])
def test_kernel_pypath(tmpdir, default_interpreter):
    """
    Test that PYTHONPATH and spyder_pythonpath option are properly handled
    when an external interpreter is used or not.

    Regression test for spyder-ide/spyder#8681.
    Regression test for spyder-ide/spyder#17511.
    """
    # Set default interpreter value
    CONF.set('main_interpreter', 'default', default_interpreter)

    # Add a path to PYTHONPATH and spyder_pythonpath config option
    pypath = to_text_string(tmpdir.mkdir('test-pypath'))
    os.environ['PYTHONPATH'] = pypath
    CONF.set('pythonpath_manager', 'spyder_pythonpath', [pypath])

    kernel_spec = SpyderKernelSpec()

    # Check that PYTHONPATH is not in our kernelspec
    # and pypath is in SPY_PYTHONPATH
    assert 'PYTHONPATH' not in kernel_spec.env
    assert pypath in kernel_spec.env['SPY_PYTHONPATH']

    # Restore default values
    CONF.set('main_interpreter', 'default', True)
    CONF.set('pythonpath_manager', 'spyder_pythonpath', [])
    del os.environ['PYTHONPATH']


def test_python_interpreter(tmpdir):
    """Test the validation of the python interpreter."""
    # Set a non existing python interpreter
    interpreter = str(tmpdir.mkdir('interpreter').join('python'))
    CONF.set('main_interpreter', 'default', False)
    CONF.set('main_interpreter', 'custom', True)
    CONF.set('main_interpreter', 'executable', interpreter)

    # Create a kernel spec
    kernel_spec = SpyderKernelSpec()

    # Assert that the python interprerter is the default one
    assert interpreter not in kernel_spec.argv
    assert CONF.get('main_interpreter', 'default')
    assert not CONF.get('main_interpreter', 'custom')


if __name__ == "__main__":
    pytest.main()



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
#

"""
Tests for the Outline explorer widgets.
"""

# Standard Libray Imports
import json
import os.path as osp
import sys
from unittest.mock import MagicMock

# Third party imports
from flaky import flaky
import pytest
from qtpy.QtCore import Qt

# Local imports
from spyder.plugins.outlineexplorer.editor import OutlineExplorerProxyEditor
from spyder.plugins.outlineexplorer.main_widget import (
    OutlineExplorerWidget, OutlineExplorerToolbuttons)
from spyder.plugins.editor.widgets.codeeditor import CodeEditor


# ---- Constants
HERE = osp.abspath(osp.dirname(__file__))
ASSETS = osp.join(HERE, 'assets')
SUFFIX = 'test_widgets'

AVAILABLE_CASES = ['text']
CASES = {
    case: {
        'file': osp.join(ASSETS, '{0}_{1}.py'.format(case, SUFFIX)),
        'data': osp.join(ASSETS, '{0}_{1}.json'.format(case, SUFFIX)),
        'tree': osp.join(ASSETS, '{0}_exp_{1}.json'.format(case, SUFFIX))
    }
    for case in AVAILABLE_CASES
}


# ---- Fixtures
@pytest.fixture
def create_outlineexplorer(qtbot):
    def _create_outlineexplorer(case, follow_cursor=False):
        case_info = CASES[case]
        filename = case_info['file']
        with open(case_info['file'], 'r') as f:
            text = f.read()

        symbol_info = json.load(open(case_info['data'], 'r'))
        expected_tree = json.load(open(case_info['tree'], 'r'))

        code_editor = CodeEditor(None)
        code_editor.set_language('py', filename)
        code_editor.show()
        code_editor.set_text(text)

        editor = OutlineExplorerProxyEditor(code_editor, filename)
        plugin_mock = MagicMock()
        plugin_mock.NAME = 'outline_explorer'

        outlineexplorer = OutlineExplorerWidget(
            'outline_explorer', plugin_mock, None)
        outlineexplorer.setup()

        outlineexplorer.set_conf('show_fullpath', True)
        outlineexplorer.set_conf('show_comments', True)
        outlineexplorer.set_conf('group_cells', True)
        outlineexplorer.set_conf('display_variables', True)
        outlineexplorer.set_conf('follow_cursor', follow_cursor)

        outlineexplorer.register_editor(editor)
        outlineexplorer.set_current_editor(editor, False, False)
        outlineexplorer.show()
        outlineexplorer.setFixedSize(400, 350)
        outlineexplorer.treewidget.is_visible = True

        editor.update_outline_info(symbol_info)
        qtbot.addWidget(outlineexplorer)
        qtbot.addWidget(code_editor)
        return outlineexplorer, expected_tree
    return _create_outlineexplorer


# ---- Tests
@pytest.mark.parametrize('case', AVAILABLE_CASES)
def test_outline_explorer(case, create_outlineexplorer):
    """
    Test to assert the outline explorer is initializing correctly and
    is showing the expected number of items, the expected type of items, and
    the expected text for each item.
    """
    outlineexplorer, expected_tree = create_outlineexplorer(case)
    assert outlineexplorer

    outlineexplorer.treewidget.expandAll()
    tree_widget = outlineexplorer.treewidget

    root_item = tree_widget.get_top_level_items()[0]
    root_ref = root_item.ref
    filename = osp.basename(root_ref.name)
    root_tree = {filename: []}
    stack = [(root_tree[filename], node) for node in root_ref.children]

    while len(stack) > 0:
        parent_tree, node = stack.pop(0)
        this_tree = {node.name: [], 'kind': node.kind}
        parent_tree.append(this_tree)
        this_stack = [(this_tree[node.name], child) for child in node.children]
        stack = this_stack + stack

    assert root_tree == expected_tree


@pytest.mark.skipif(sys.platform == 'darwin', reason="Fails on Mac")
def test_go_to_cursor_position(create_outlineexplorer, qtbot):
    """
    Test that clicking on the 'Go to cursor position' button located in the
    toolbar of the outline explorer is working as expected.

    Regression test for spyder-ide/spyder#7729.
    """
    outlineexplorer, _ = create_outlineexplorer('text')

    # Move the mouse cursor in the editor to line 15
    editor = outlineexplorer.treewidget.current_editor
    editor._editor.go_to_line(15)
    assert editor._editor.get_text_line(14) == "    def inner():"

    # Click on the 'Go to cursor position' button of the outline explorer's
    # toolbar :
    assert outlineexplorer.treewidget.currentItem() is None
    qtbot.mouseClick(
        outlineexplorer.get_toolbutton(OutlineExplorerToolbuttons.GoToCursor),
        Qt.LeftButton)
    assert outlineexplorer.treewidget.currentItem().text(0) == 'inner'


@flaky(max_runs=10)
def test_follow_cursor(create_outlineexplorer, qtbot):
    """
    Test that the cursor is followed.
    """
    outlineexplorer, _ = create_outlineexplorer('text', follow_cursor=True)

    # Move the mouse cursor in the editor to line 52
    editor = outlineexplorer.treewidget.current_editor
    editor._editor.go_to_line(52)
    assert editor._editor.get_text_line(51) == \
           "        super(Class1, self).__init__()"

    # __init__ is collapsed
    assert outlineexplorer.treewidget.currentItem().text(0) == '__init__'

    # Go to cursor to open the cursor
    qtbot.mouseClick(
        outlineexplorer.get_toolbutton(OutlineExplorerToolbuttons.GoToCursor),
        Qt.LeftButton)

    # Check if follows
    editor._editor.go_to_line(1)
    text = outlineexplorer.treewidget.currentItem().text(0)
    assert text == CASES['text']['file']

    editor._editor.go_to_line(37)
    assert editor._editor.get_text_line(36) == "# %%%% cell level 3"
    assert outlineexplorer.treewidget.currentItem().text(0) == 'cell level 3'


@flaky(max_runs=10)
def test_go_to_cursor_position_with_new_file(create_outlineexplorer, qtbot):
    """
    Test that clicking on the 'Go to cursor position' button located in the
    toolbar of the outline explorer is working as expected for newly created
    files.

    Regression test for spyder-ide/spyder#8510.
    """
    # text = "# -*- coding: utf-8 -*-\nSome newly created\nPython file."
    outlineexplorer, _ = create_outlineexplorer('text')

    # Click on the 'Go to cursor position' button of the outline explorer's
    # toolbar :
    filename = CASES['text']['file']
    qtbot.mouseClick(
        outlineexplorer.get_toolbutton(OutlineExplorerToolbuttons.GoToCursor),
        Qt.LeftButton)
    assert outlineexplorer.treewidget.currentItem().text(0) == filename


@flaky(max_runs=10)
def test_go_to_last_item(create_outlineexplorer, qtbot):
    """
    Test that clicking on the 'Go to cursor position' button located in the
    toolbar of the outline explorer is working as expected when the cursor
    is located in the editor under the last item of the outline tree widget.

    Regression test for spyder-ide/spyder#7744.
    """
    outlineexplorer, _ = create_outlineexplorer('text')

    # Move the mouse cursor in the editor to the last line :
    editor = outlineexplorer.treewidget.current_editor
    line_count = editor._editor.document().blockCount() - 2
    editor._editor.go_to_line(line_count)
    assert editor._editor.get_text_line(line_count) == "        pass"

    # Click on the 'Go to cursor position' button of the outline explorer's
    # toolbar :
    qtbot.mouseClick(
        outlineexplorer.get_toolbutton(OutlineExplorerToolbuttons.GoToCursor),
        Qt.LeftButton)
    assert outlineexplorer.treewidget.currentItem().text(0) == 'method1'


@flaky(max_runs=10)
def test_display_variables(create_outlineexplorer, qtbot):
    """
    Test that clicking on the 'Display variables and attributes' button located in the
    toolbar of the outline explorer is working as expected by updating the tree widget.

    Regression test for spyder-ide/spyder#21456.
    """
    outlineexplorer, _ = create_outlineexplorer('text')

    editor = outlineexplorer.treewidget.current_editor
    state = outlineexplorer.treewidget.display_variables

    editor_id = editor.get_id()

    initial_tree = outlineexplorer.treewidget.editor_tree_cache[editor_id]

    outlineexplorer.treewidget.toggle_variables(not state)

    first_toggle_tree = outlineexplorer.treewidget.editor_tree_cache[editor_id]

    assert first_toggle_tree != initial_tree

    outlineexplorer.treewidget.toggle_variables(state)

    second_toggle_tree = outlineexplorer.treewidget.editor_tree_cache[editor_id]

    assert (second_toggle_tree != first_toggle_tree) and (
        second_toggle_tree == initial_tree)


if __name__ == "__main__":
    import os
    pytest.main(['-x', os.path.basename(__file__), '-v', '-rw'])



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""
Spyder application menu constants.
"""

# Local imports
from spyder.api.widgets.menus import SpyderMenu
from spyder.utils.palette import SpyderPalette
from spyder.utils.stylesheet import AppStyle, SpyderStyleSheet


# ---- Stylesheet
# -----------------------------------------------------------------------------
class _MenuBarStylesheet(SpyderStyleSheet):
    """Stylesheet for menubars used in Spyder."""

    def set_stylesheet(self):
        css = self.get_stylesheet()

        # Set the same color as the one used for the app toolbar
        css.QMenuBar.setValues(
            backgroundColor=SpyderPalette.COLOR_BACKGROUND_4
        )

        # Give more padding and margin to items
        css['QMenuBar::item'].setValues(
            padding=f'{2 * AppStyle.MarginSize}px',
            margin='0px 2px'
        )

        # Remove padding when pressing main menus
        css['QMenuBar::item:pressed'].setValues(
            padding='0px'
        )

        # Set hover and pressed state of items in the menu bar
        for state in ['selected', 'pressed']:
            # Don't use a different color for the QMenuBar pressed state
            # because a lighter color has too little contrast with the text.
            bg_color = SpyderPalette.COLOR_BACKGROUND_5

            css[f"QMenuBar::item:{state}"].setValues(
                backgroundColor=bg_color,
                borderRadius=SpyderPalette.SIZE_BORDER_RADIUS
            )


MENUBAR_STYLESHEET = _MenuBarStylesheet()


# ---- Menu constants
# -----------------------------------------------------------------------------
class ApplicationContextMenu:
    Documentation = 'context_documentation_section'
    About = 'context_about_section'


class ApplicationMenus:
    File = 'file_menu'
    Edit = 'edit_menu'
    Search = 'search_menu'
    Source = 'source_menu'
    Run = 'run_menu'
    Debug = 'debug_menu'
    Consoles = 'consoles_menu'
    Projects = 'projects_menu'
    Tools = 'tools_menu'
    View = 'view_menu'
    Help = 'help_menu'


class FileMenuSections:
    New = 'new_section'
    Open = 'open_section'
    Save = 'save_section'
    Print = 'print_section'
    Close = 'close_section'
    Switcher = 'switcher_section'
    Navigation = 'navigation_section'
    Restart = 'restart_section'


class EditMenuSections:
    UndoRedo = 'undo_redo_section'
    Copy = 'copy_section'
    Editor = 'editor_section'


class SearchMenuSections:
    FindInText = 'find_text_section'
    FindInFiles = 'find_files_section'


class SourceMenuSections:
    Options = 'options_section'
    Linting = 'linting_section'
    Cursor = 'cursor_section'
    Formatting = 'formatting_section'
    CodeAnalysis = 'code_analysis_section'


class RunMenuSections:
    Run = 'run_section'
    RunExtras = 'run_extras_section'
    RunInExecutors = 'executors_section'

class DebugMenuSections:
    StartDebug = 'start_debug_section'
    ControlDebug = 'control_debug_section'
    EditBreakpoints = 'edit_breakpoints_section'


class ConsolesMenuSections:
    New = 'new_section'
    Restart = 'restart_section'


class ProjectsMenuSections:
    New = 'new_section'
    Open = 'open_section'
    Extras = 'extras_section'


class ToolsMenuSections:
    Tools = 'tools_section'
    External = 'external_section'
    Extras = 'extras_section'


class ViewMenuSections:
    Top = 'top_section'
    Pane = 'pane_section'
    Toolbar = 'toolbar_section'
    Layout = 'layout_section'
    Bottom = 'bottom_section'


class HelpMenuSections:
    Documentation = 'documentation_section'
    Support = 'support_section'
    ExternalDocumentation = 'external_documentation_section'
    About = 'about_section'


# ---- App menu class
# -----------------------------------------------------------------------------
class ApplicationMenu(SpyderMenu):
    """
    Spyder main window application menu.

    This class provides application menus with some predefined functionality.
    """

    APP_MENU = True



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) Spyder Project Contributors
#
# Licensed under the terms of the MIT License
# (see LICENSE.txt for details)
# -----------------------------------------------------------------------------

# Standard library imports
import sys
import time

# Third-party imports
import pytest

# Local imports
from spyder.api.plugins import Plugins
from spyder.api.plugin_registration.registry import PLUGIN_REGISTRY
from spyder.plugins.maininterpreter.plugin import MainInterpreter
from spyder.plugins.preferences.tests.conftest import MainWindowMock
from spyder.utils.conda import get_list_conda_envs
from spyder.utils.pyenv import get_list_pyenv_envs


# Get envs to show them in the Main interpreter page. This is actually
# done in a thread in the MainInterpreter container.
# We're also recording the time needed to get them to compare it with the
# loading time of that config page.
t0 = time.time()
conda_envs = get_list_conda_envs()
pyenv_envs = get_list_pyenv_envs()
GET_ENVS_TIME = time.time() - t0


@pytest.mark.skipif(
    ((len(conda_envs) == 0 and len(pyenv_envs) == 0) or
     sys.platform == 'darwin'),
    reason="Makes no sense if conda and pyenv are not installed, fails on mac"
)
def test_load_time(qtbot):
    # Create Preferences dialog
    main = MainWindowMock(None)
    preferences = main.get_plugin(Plugins.Preferences)

    PLUGIN_REGISTRY.register_plugin(main, MainInterpreter)

    # Create page and measure time to do it
    t0 = time.time()
    preferences.open_dialog()
    load_time = time.time() - t0

    container = preferences.get_container()
    dlg = container.dialog
    widget = dlg.get_page()

    # Assert the combobox is populated with the found envs
    assert widget.cus_exec_combo.combobox.count() > 0

    # Assert load time is smaller than the one required to get envs
    # directly. This means we're using the cached envs instead
    assert load_time < GET_ENVS_TIME

    # Load time should be small too because we perform simple validations
    # on the page.
    assert load_time < 0.5



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
#
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""Remote client container."""

# Standard library imports
from __future__ import annotations
from collections import deque
import functools

# Third-party imports
from qtpy.QtCore import Signal
from qtpy.QtWidgets import QMessageBox

# Local imports
from spyder.api.translations import _
from spyder.api.widgets.main_container import PluginMainContainer
from spyder.plugins.ipythonconsole.utils.kernel_handler import KernelHandler
from spyder.plugins.remoteclient import SPYDER_REMOTE_MAX_VERSION
from spyder.plugins.remoteclient.api import (
    MAX_CLIENT_MESSAGES,
    RemoteClientActions,
    RemoteClientMenus,
    RemoteConsolesMenuSections,
)
from spyder.plugins.remoteclient.api.protocol import ConnectionInfo
from spyder.plugins.remoteclient.widgets.connectiondialog import (
    ConnectionDialog,
)


class RemoteClientContainer(PluginMainContainer):

    _sig_kernel_restarted = Signal(object, bool)
    """
    This private signal is used to inform that a kernel restart took place in
    the server.

    Parameters
    ----------
    ipyclient: ClientWidget
        An IPython console client widget (the first parameter in both
        signatures).
    response: bool or None
        Response returned by the server. `None` can happen when the connection
        to the server is lost.
    """

    _sig_kernel_info_replied = Signal(object, dict)
    """
    This private signal is used to inform that a kernel info request to the
    server was replied.

    Parameters
    ----------
    ipyclient: ClientWidget
        An IPython console client widget (the first parameter in both
        signatures).
    response: dict or None
        Response returned by the server. `None` can happen when the connection
        to the server is lost.
    """

    sig_start_server_requested = Signal(str)
    """
    This signal is used to request starting a remote server.

    Parameters
    ----------
    id: str
        Id of the server that will be started.
    """

    sig_stop_server_requested = Signal(str)
    """
    This signal is used to request stopping a remote server.

    Parameters
    ----------
    id: str
        Id of the server that will be stopped.
    """

    sig_server_renamed = Signal(str)
    """
    This signal is used to inform that a remote server was renamed.

    Parameters
    ----------
    id: str
        Id of the server that was renamed.
    """

    sig_connection_status_changed = Signal(dict)
    """
    This signal is used to update the status of a given connection.

    Parameters
    ----------
    info: ConnectionInfo
        Dictionary with the necessary info to update the status of a
        connection.
    """

    sig_create_ipyclient_requested = Signal(str)
    """
    This signal is used to request starting an IPython console client for a
    remote server.

    Parameters
    ----------
    id: str
        Id of the server for which a client will be created.
    """

    sig_shutdown_kernel_requested = Signal(str, str)
    """
    This signal is used to request shutting down a kernel.

    Parameters
    ----------
    id: str
        Id of the server for which a kernel shutdown will be requested.
    kernel_id: str
        Id of the kernel which will be shutdown in the server.
    """

    sig_interrupt_kernel_requested = Signal(str, str)
    """
    This signal is used to request interrupting a kernel.

    Parameters
    ----------
    id: str
        Id of the server for which a kernel interrupt will be requested.
    kernel_id: str
        Id of the kernel which will be shutdown in the server.
    """

    sig_client_message_logged = Signal(dict)
    """
    This signal is used to inform that a client has logged a connection
    message.

    Parameters
    ----------
    log: RemoteClientLog
        Dictionary that contains the log message and its metadata.
    """

    # ---- PluginMainContainer API
    # -------------------------------------------------------------------------
    def setup(self):
        # Attributes
        self.client_logs: dict[str, deque] = {}

        # Widgets
        self.create_action(
            RemoteClientActions.ManageConnections,
            _("Manage remote connections..."),
            icon=self._plugin.get_icon(),
            triggered=self._show_connection_dialog,
        )

        self._remote_consoles_menu = self.create_menu(
            RemoteClientMenus.RemoteConsoles, _("New console in remote server")
        )

        # Signals
        self.sig_connection_status_changed.connect(
            self._on_connection_status_changed
        )
        self.sig_client_message_logged.connect(self._on_client_message_logged)
        self._sig_kernel_restarted.connect(self._on_kernel_restarted)
        self._sig_kernel_info_replied.connect(self._on_kernel_info_reply)

        self.__requested_restart = False
        self.__requested_info = False

    def update_actions(self):
        pass

    # ---- Public API
    # -------------------------------------------------------------------------
    def setup_remote_consoles_submenu(self, render=True):
        """Create the remote consoles submenu in the Consoles app one."""
        self._remote_consoles_menu.clear_actions()

        self.add_item_to_menu(
            self.get_action(RemoteClientActions.ManageConnections),
            menu=self._remote_consoles_menu,
            section=RemoteConsolesMenuSections.ManagerSection,
        )

        servers = self.get_conf("servers", default={})
        for config_id in servers:
            auth_method = self.get_conf(f"{config_id}/auth_method")
            name = self.get_conf(f"{config_id}/{auth_method}/name")

            action = self.create_action(
                name=config_id,
                text=f"New console in {name} server",
                icon=self.create_icon("ipython_console"),
                triggered=functools.partial(
                    self.sig_create_ipyclient_requested.emit,
                    config_id,
                ),
                overwrite=True,
            )
            self.add_item_to_menu(
                action,
                menu=self._remote_consoles_menu,
                section=RemoteConsolesMenuSections.ConsolesSection,
            )

        # This is necessary to reposition the menu correctly when rebuilt
        if render:
            self._remote_consoles_menu.render()

    def on_kernel_started(self, ipyclient, kernel_info):
        """
        Actions to take when a remote kernel was started for an IPython console
        client.
        """
        config_id = ipyclient.server_id

        # It's only at this point that we can allow users to close the client.
        ipyclient.can_close = True

        # Handle failures to launch a kernel
        if not kernel_info:
            auth_method = self.get_conf(f"{config_id}/auth_method")
            name = self.get_conf(f"{config_id}/{auth_method}/name")
            ipyclient.show_kernel_error(
                _(
                    "There was an error connecting to the server <b>{}</b>. "
                    "Please check your connection is working."
                ).format(name)
            )
            return

        # Connect client's signals
        ipyclient.kernel_id = kernel_info["id"]
        self._connect_ipyclient_signals(ipyclient)

        try:
            kernel_handler = KernelHandler.from_connection_info(
                kernel_info["connection_info"],
                ssh_connection=self._plugin.get_remote_server(
                    ipyclient.server_id
                )._ssh_connection,
            )

            # Need to be smaller than the usual time it takes for the kernel to
            # restart
            kernel_handler.set_time_to_dead(1.0)
        except Exception as err:
            ipyclient.show_kernel_error(err)
        else:
            # Connect client to the kernel
            ipyclient.connect_kernel(kernel_handler)

    def on_server_version_mismatch(self, config_id, version: str):
        """
        Actions to take when there's a mismatch between the
        spyder-remote-services version installed in the server and the one
        supported by Spyder.
        """
        auth_method = self.get_conf(f"{config_id}/auth_method")
        server_name = self.get_conf(f"{config_id}/{auth_method}/name")

        QMessageBox.critical(
            self,
            _("Remote server error"),
            _(
                "The version of <tt>spyder-remote-services</tt> on the "
                "remote host <b>{server}</b> (<b>{srs_version}</b>) is newer "
                "than the latest Spyder supports (<b>{max_version}</b>)."
                "<br><br>"
                "Please update Spyder to be able to connect to this host."
            ).format(
                server=server_name,
                srs_version=version,
                max_version=SPYDER_REMOTE_MAX_VERSION,
            ),
            QMessageBox.Ok,
        )

    # ---- Private API
    # -------------------------------------------------------------------------
    def _show_connection_dialog(self):
        connection_dialog = ConnectionDialog(self)

        connection_dialog.sig_start_server_requested.connect(
            self.sig_start_server_requested
        )
        connection_dialog.sig_stop_server_requested.connect(
            self.sig_stop_server_requested
        )
        connection_dialog.sig_connections_changed.connect(
            self.setup_remote_consoles_submenu
        )
        connection_dialog.sig_server_renamed.connect(self.sig_server_renamed)

        connection_dialog.show()

    def _on_connection_status_changed(self, info: ConnectionInfo):
        """Handle changes in connection status."""
        host_id = info["id"]
        status = info["status"]
        message = info["message"]

        # We need to save this info so that we can show the current status in
        # the connection dialog when it's closed and opened again.
        self.set_conf(f"{host_id}/status", status)
        self.set_conf(f"{host_id}/status_message", message)

    def _connect_ipyclient_signals(self, ipyclient):
        """
        Connect the signals to shutdown, interrupt and restart the kernel of an
        IPython console client to the signals and methods declared here for
        remote kernel management.
        """
        ipyclient.sig_shutdown_kernel_requested.connect(
            self.sig_shutdown_kernel_requested
        )
        ipyclient.sig_interrupt_kernel_requested.connect(
            self.sig_interrupt_kernel_requested
        )
        ipyclient.sig_restart_kernel_requested.connect(
            lambda: self._request_kernel_restart(ipyclient)
        )
        ipyclient.sig_kernel_died.connect(
            lambda: self._request_kernel_info(ipyclient)
        )

    def _request_kernel_restart(self, ipyclient):
        """
        Request a kernel restart to the server for an IPython console client
        and handle its response.
        """
        if self.__requested_restart:
            return

        future = self._plugin._restart_kernel(
            ipyclient.server_id, ipyclient.kernel_id
        )

        self.__requested_restart = True

        future.add_done_callback(
            lambda future: self._sig_kernel_restarted.emit(
                ipyclient, future.result()
            )
        )

    def _request_kernel_info(self, ipyclient):
        """
        Request a kernel reconnect to the server for an IPython console client
        and handle its response.
        """
        if self.__requested_info:
            return

        future = self._plugin._get_kernel_info(
            ipyclient.server_id, ipyclient.kernel_id
        )

        self.__requested_info = True

        future.add_done_callback(
            lambda future: self._sig_kernel_info_replied.emit(
                ipyclient, future.result()
            )
        )

    def _on_kernel_restarted(self, ipyclient, restarted: bool):
        """
        Get kernel info corresponding to an IPython console client from the
        server.

        If we get a response, it means the kernel is alive.
        """
        self.__requested_restart = False
        if restarted:
            ipyclient.kernel_handler.reconnect_kernel()
            ipyclient.handle_remote_kernel_restarted()
        else:
            ipyclient.remote_kernel_restarted_failure_message(shutdown=True)

    def _on_kernel_info_reply(self, ipyclient, kernel_info):
        """Check spyder-kernels version."""
        self.__requested_info = False
        if kernel_info:
            try:
                kernel_handler = KernelHandler.from_connection_info(
                    kernel_info["connection_info"],
                    ssh_connection=self._plugin.get_remote_server(
                        ipyclient.server_id
                    )._ssh_connection,
                )
                kernel_handler.set_time_to_dead(1.0)
            except Exception as err:
                ipyclient.remote_kernel_restarted_failure_message(
                    err, shutdown=True
                )
            else:
                ipyclient.replace_kernel(
                    kernel_handler, shutdown_kernel=False, clear=False
                )
        else:
            ipyclient.remote_kernel_restarted_failure_message(shutdown=True)

            # This will show an error message in the plugins connected to the
            # IPython console and disable kernel related actions in its Options
            # menu.
            sw = ipyclient.shellwidget
            sw.sig_shellwidget_errored.emit(sw)

    def _on_client_message_logged(self, message: dict):
        """Actions to take when a client message is logged."""
        msg_id = message["id"]

        # Create deque if not available
        if not self.client_logs.get(msg_id):
            self.client_logs[msg_id] = deque([], MAX_CLIENT_MESSAGES)

        # Add message to deque
        self.client_logs[msg_id].append(message)



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""Watcher to detect filesystem changes in the project's directory."""

# Standard lib imports
import os
import logging
from pathlib import Path

# Third-party imports
from qtpy.QtCore import QObject, Signal
from superqt.utils import qthrottled
import watchdog
from watchdog.events import FileSystemEventHandler, PatternMatchingEventHandler
from watchdog.observers.polling import PollingObserverVFS

# Local imports
from spyder.config.utils import EDIT_EXTENSIONS


# ---- Constants
# -----------------------------------------------------------------------------
logger = logging.getLogger(__name__)

FOLDERS_TO_IGNORE = [
    "__pycache__",
    "build",
]


# ---- Monkey patches
# -----------------------------------------------------------------------------
class BaseThreadWrapper(watchdog.utils.BaseThread):
    """
    Wrapper around watchdog BaseThread class.
    This is necessary for issue spyder-ide/spyder#11370
    """
    queue = None

    def __init__(self):
        super(BaseThreadWrapper, self).__init__()
        self._original_run = self.run
        self.run = self.run_wrapper

    def run_wrapper(self):
        try:
            self._original_run()
        except OSError as e:
            logger.exception('Watchdog thread exited with error %s',
                             e.strerror)
            self.queue.put(e)


# Monkeypatching BaseThread to prevent the error reported in
# spyder-ide/spyder#11370
watchdog.utils.BaseThread = BaseThreadWrapper


# ---- Auxiliary functions
# -----------------------------------------------------------------------------
def ignore_entry(entry: os.DirEntry) -> bool:
    """Check if an entry should be ignored."""
    parts = Path(entry.path).parts

    # Ignore files in hidden directories (e.g. .git)
    if any([p.startswith(".") for p in parts]):
        return True

    # Ignore specific folders
    for folder in FOLDERS_TO_IGNORE:
        if folder in parts:
            return True

    return False


def editable_file(entry: os.DirEntry) -> bool:
    """Check if an entry file is editable."""
    if entry.is_file():
        return (os.path.splitext(entry.path)[1] in EDIT_EXTENSIONS)
    return True


def filter_scandir(path):
    """
    Filter entries from os.scandir that we're not interested in tracking in the
    observer.
    """
    return (
        entry for entry in os.scandir(path)
        if (not ignore_entry(entry) and editable_file(entry))
    )


# ---- Event handler
# -----------------------------------------------------------------------------
class WorkspaceEventHandler(QObject, PatternMatchingEventHandler):
    """
    Event handler for watchdog notifications.

    This class receives notifications about file/folder moving, modification,
    creation and deletion and emits a corresponding signal about it.
    """

    sig_file_moved = Signal(str, str, bool)
    sig_file_created = Signal(str, bool)
    sig_file_deleted = Signal(str, bool)
    sig_file_modified = Signal(str, bool)

    def __init__(self, parent=None):
        QObject.__init__(self, parent)
        PatternMatchingEventHandler.__init__(
            self,
            patterns=[f"*{ext}" for ext in EDIT_EXTENSIONS],
        )

    def fmt_is_dir(self, is_dir):
        return 'directory' if is_dir else 'file'

    def on_moved(self, event):
        src_path = event.src_path
        dest_path = event.dest_path
        is_dir = event.is_directory
        logger.info("Moved {0}: {1} to {2}".format(
            self.fmt_is_dir(is_dir), src_path, dest_path))
        self.sig_file_moved.emit(src_path, dest_path, is_dir)

    def on_created(self, event):
        src_path = event.src_path
        is_dir = event.is_directory
        logger.info("Created {0}: {1}".format(
            self.fmt_is_dir(is_dir), src_path))
        self.sig_file_created.emit(src_path, is_dir)

    def on_deleted(self, event):
        src_path = event.src_path
        is_dir = event.is_directory
        logger.info("Deleted {0}: {1}".format(
            self.fmt_is_dir(is_dir), src_path))
        self.sig_file_deleted.emit(src_path, is_dir)

    def on_modified(self, event):
        src_path = event.src_path
        is_dir = event.is_directory
        logger.info("Modified {0}: {1}".format(
            self.fmt_is_dir(is_dir), src_path))
        self.sig_file_modified.emit(src_path, is_dir)

    def dispatch(self, event):
        # Don't apply patterns to directories, only to files
        if event.is_directory:
            FileSystemEventHandler.dispatch(self, event)
        else:
            super().dispatch(event)


# ---- Watcher
# -----------------------------------------------------------------------------
class WorkspaceWatcher(QObject):
    """
    Wrapper class around watchdog observer and notifier.

    It provides methods to start and stop watching folders.
    """

    observer = None

    sig_file_moved = Signal(str, str, bool)
    sig_file_created = Signal(str, bool)
    sig_file_deleted = Signal(str, bool)
    sig_file_modified = Signal(str, bool)

    def __init__(self, parent=None):
        super().__init__(parent)
        self.event_handler = WorkspaceEventHandler(self)

        self.event_handler.sig_file_moved.connect(self.on_moved)
        self.event_handler.sig_file_created.connect(self.on_created)
        self.event_handler.sig_file_deleted.connect(self.on_deleted)
        self.event_handler.sig_file_modified.connect(self.on_modified)

    def connect_signals(self, project):
        self.sig_file_created.connect(project.file_created)
        self.sig_file_moved.connect(project.file_moved)
        self.sig_file_deleted.connect(project.file_deleted)
        self.sig_file_modified.connect(project.file_modified)

    def start(self, workspace_folder):
        # We use a polling observer because:
        # * It doesn't introduce long freezes on Linux when switching git
        #   branches that have many changes between them. That's because the
        #   OS-based observer (i.e. inotify) generates way too many events.
        # * The OS-based observer on Windows has many shortcomings (see
        #   openmsi/openmsistream#56).
        # * There doesn't seem to be issues on Mac, but it's simpler to use a
        #   single observer for all OSes.
        self.observer = PollingObserverVFS(
            stat=os.stat, listdir=filter_scandir
        )

        self.observer.schedule(
            self.event_handler, workspace_folder, recursive=True
        )

        try:
            self.observer.start()
        except Exception:
            logger.debug(
                f"Observer could not be started for: {workspace_folder}."
            )

    def stop(self):
        if self.observer is not None:
            # This is required to avoid showing an error when closing
            # projects.
            # Fixes spyder-ide/spyder#14107
            try:
                self.observer.stop()
                self.observer.join()
                del self.observer
                self.observer = None
            except RuntimeError:
                pass

    @qthrottled(timeout=200)
    def on_moved(self, src_path, dest_path, is_dir):
        self.sig_file_moved.emit(src_path, dest_path, is_dir)

    @qthrottled(timeout=200)
    def on_created(self, path, is_dir):
        self.sig_file_created.emit(path, is_dir)

    @qthrottled(timeout=200)
    def on_deleted(self, path, is_dir):
        self.sig_file_deleted.emit(path, is_dir)

    @qthrottled(timeout=200)
    def on_modified(self, path, is_dir):
        self.sig_file_modified.emit(path, is_dir)



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""
Projects Plugin

It handles closing, opening and switching among projects and also
updating the file tree explorer associated with a project.
"""

# Standard library imports
import logging
import os.path as osp

# Third party imports
from qtpy.QtCore import Signal

# Local imports
from spyder.api.plugin_registration.decorators import (
    on_plugin_available, on_plugin_teardown)
from spyder.api.plugins import Plugins, SpyderDockablePlugin
from spyder.api.translations import _
from spyder.plugins.completion.api import WorkspaceUpdateKind
from spyder.plugins.mainmenu.api import ApplicationMenus, ProjectsMenuSections
from spyder.plugins.projects.api import EmptyProject
from spyder.plugins.projects.widgets.main_widget import (
    ProjectsActions, ProjectExplorerWidget)
from spyder.utils.misc import getcwd_or_home


# Logging
logger = logging.getLogger(__name__)


class Projects(SpyderDockablePlugin):
    """Projects plugin."""

    NAME = 'project_explorer'
    CONF_SECTION = NAME
    CONF_FILE = False
    REQUIRES = []
    OPTIONAL = [Plugins.Completions, Plugins.IPythonConsole, Plugins.Editor,
                Plugins.MainMenu, Plugins.Switcher]
    WIDGET_CLASS = ProjectExplorerWidget

    # Signals
    sig_project_created = Signal(str, str)
    """
    This signal is emitted to request the Projects plugin the creation of a
    project.

    Parameters
    ----------
    project_path: str
        Location of project.
    project_type: str
        Type of project as defined by project types.
    """

    sig_project_loaded = Signal(str)
    """
    This signal is emitted when a project is loaded.

    Parameters
    ----------
    project_path: str
        Loaded project path.
    """

    sig_project_closed = Signal((str,), (bool,))
    """
    This signal is emitted when a project is closed.

    Parameters
    ----------
    project_path: str
        Closed project path (signature 1).
    close_project: bool
        This is emitted only when closing a project but not when switching
        between projects (signature 2).
    """

    # ---- SpyderDockablePlugin API
    # -------------------------------------------------------------------------
    @staticmethod
    def get_name():
        return _("Projects")

    @staticmethod
    def get_description():
        return _("Create Spyder projects and manage their files.")

    @classmethod
    def get_icon(cls):
        return cls.create_icon('project_spyder')

    def on_initialize(self):
        """Register plugin in Spyder's main window"""
        widget = self.get_widget()
        treewidget = widget.treewidget
        self._completions = None
        self._switcher = None

        # Emit public signals so that other plugins can connect to them
        widget.sig_project_created.connect(self.sig_project_created)
        widget.sig_project_closed.connect(self.sig_project_closed)
        widget.sig_project_loaded.connect(self.sig_project_loaded)

        treewidget.sig_delete_project.connect(self.delete_project)
        treewidget.sig_redirect_stdio_requested.connect(
            self.sig_redirect_stdio_requested)
        self.sig_switch_to_plugin_requested.connect(
            lambda plugin, check: self._show_main_widget())

        if self.main:
            widget.sig_open_file_requested.connect(self.main.open_file)
            widget.sig_project_loaded.connect(
                lambda v: self.main.set_window_title())
            widget.sig_project_closed.connect(
                lambda v: self.main.set_window_title())
            self.main.restore_scrollbar_position.connect(
                self.get_widget().restore_scrollbar_position)

        self.register_project_type(self, EmptyProject)

    @on_plugin_available(plugin=Plugins.Editor)
    def on_editor_available(self):
        editor = self.get_plugin(Plugins.Editor)
        widget = self.get_widget()
        treewidget = widget.treewidget

        treewidget.sig_open_file_requested.connect(editor.load)
        treewidget.sig_removed.connect(editor.removed)
        treewidget.sig_tree_removed.connect(editor.removed_tree)
        treewidget.sig_renamed.connect(editor.renamed)
        treewidget.sig_tree_renamed.connect(editor.renamed_tree)
        treewidget.sig_module_created.connect(editor.new)
        treewidget.sig_file_created.connect(self._new_editor)

        widget.sig_save_open_files_requested.connect(editor.save_open_files)
        widget.sig_project_loaded.connect(self._setup_editor_files)
        widget.sig_project_closed[bool].connect(self._setup_editor_files)
        widget.sig_project_loaded.connect(self._set_path_in_editor)
        widget.sig_project_closed.connect(self._unset_path_in_editor)
        # To handle switcher open request
        widget.sig_open_file_requested.connect(editor.load)

    @on_plugin_available(plugin=Plugins.Completions)
    def on_completions_available(self):
        self._completions = self.get_plugin(Plugins.Completions)
        widget = self.get_widget()

        # TODO: This is not necessary anymore due to us starting workspace
        # services in the editor. However, we could restore it in the future.
        # completions.sig_language_completions_available.connect(
        #     lambda settings, language:
        #         self.start_workspace_services())
        self._completions.sig_stop_completions.connect(
            self.stop_workspace_services)
        widget.sig_project_loaded.connect(self._add_path_to_completions)
        widget.sig_project_closed.connect(self._remove_path_from_completions)
        widget.sig_broadcast_notification_requested.connect(
            self._broadcast_notification)

    @on_plugin_available(plugin=Plugins.IPythonConsole)
    def on_ipython_console_available(self):
        ipyconsole = self.get_plugin(Plugins.IPythonConsole)
        widget = self.get_widget()
        treewidget = widget.treewidget

        widget.sig_restart_console_requested.connect(ipyconsole.restart)
        treewidget.sig_open_interpreter_requested.connect(
            ipyconsole.create_client_from_path)
        treewidget.sig_run_requested.connect(self._run_file_in_ipyconsole)

    @on_plugin_available(plugin=Plugins.MainMenu)
    def on_main_menu_available(self):
        main_menu = self.get_plugin(Plugins.MainMenu)
        new_project_action = self.get_action(ProjectsActions.NewProject)
        open_project_action = self.get_action(ProjectsActions.OpenProject)
        close_project_action = self.get_action(ProjectsActions.CloseProject)
        delete_project_action = self.get_action(ProjectsActions.DeleteProject)

        projects_menu = main_menu.get_application_menu(
            ApplicationMenus.Projects)
        projects_menu.aboutToShow.connect(self._is_invalid_active_project)

        main_menu.add_item_to_application_menu(
            new_project_action,
            menu_id=ApplicationMenus.Projects,
            section=ProjectsMenuSections.New)

        for item in [open_project_action, close_project_action,
                     delete_project_action]:
            main_menu.add_item_to_application_menu(
                item,
                menu_id=ApplicationMenus.Projects,
                section=ProjectsMenuSections.Open)

        main_menu.add_item_to_application_menu(
            self.get_widget().recent_project_menu,
            menu_id=ApplicationMenus.Projects,
            section=ProjectsMenuSections.Extras)

    @on_plugin_available(plugin=Plugins.Switcher)
    def on_switcher_available(self):
        # Connect to switcher
        self._switcher = self.get_plugin(Plugins.Switcher)
        self._switcher.sig_mode_selected.connect(self._handle_switcher_modes)
        self._switcher.sig_item_selected.connect(
            self._handle_switcher_selection)
        self._switcher.sig_search_text_available.connect(
            self._handle_switcher_search)

    @on_plugin_teardown(plugin=Plugins.Editor)
    def on_editor_teardown(self):
        editor = self.get_plugin(Plugins.Editor)
        widget = self.get_widget()
        treewidget = widget.treewidget

        treewidget.sig_open_file_requested.disconnect(editor.load)
        treewidget.sig_removed.disconnect(editor.removed)
        treewidget.sig_tree_removed.disconnect(editor.removed_tree)
        treewidget.sig_renamed.disconnect(editor.renamed)
        treewidget.sig_tree_renamed.disconnect(editor.renamed_tree)
        treewidget.sig_module_created.disconnect(editor.new)
        treewidget.sig_file_created.disconnect(self._new_editor)

        widget.sig_save_open_files_requested.disconnect(editor.save_open_files)
        widget.sig_project_loaded.disconnect(self._setup_editor_files)
        widget.sig_project_closed[bool].disconnect(self._setup_editor_files)
        widget.sig_project_loaded.disconnect(self._set_path_in_editor)
        widget.sig_project_closed.disconnect(self._unset_path_in_editor)
        # To handle switcher open request
        widget.sig_open_file_requested.disconnect(editor.load)

    @on_plugin_teardown(plugin=Plugins.Completions)
    def on_completions_teardown(self):
        self._completions = self.get_plugin(Plugins.Completions)
        widget = self.get_widget()

        self._completions.sig_stop_completions.disconnect(
            self.stop_workspace_services)

        widget.sig_project_loaded.disconnect(self._add_path_to_completions)
        widget.sig_project_closed.disconnect(
            self._remove_path_from_completions)
        widget.sig_broadcast_notification_requested.disconnect(
            self._broadcast_notification)

        self._completions = None

    @on_plugin_teardown(plugin=Plugins.IPythonConsole)
    def on_ipython_console_teardown(self):
        ipyconsole = self.get_plugin(Plugins.IPythonConsole)
        widget = self.get_widget()
        treewidget = widget.treewidget

        widget.sig_restart_console_requested.disconnect(ipyconsole.restart)
        treewidget.sig_open_interpreter_requested.disconnect(
            ipyconsole.create_client_from_path)
        treewidget.sig_run_requested.disconnect(self._run_file_in_ipyconsole)

    @on_plugin_teardown(plugin=Plugins.MainMenu)
    def on_main_menu_teardown(self):
        main_menu = self.get_plugin(Plugins.MainMenu)
        main_menu.remove_application_menu(ApplicationMenus.Projects)

    @on_plugin_teardown(plugin=Plugins.Switcher)
    def on_switcher_teardown(self):
        # Disconnect from switcher
        self._switcher.sig_mode_selected.disconnect(
            self._handle_switcher_modes)
        self._switcher.sig_item_selected.disconnect(
            self._handle_switcher_selection)
        self._switcher.sig_search_text_available.disconnect(
            self._handle_switcher_search)
        self._switcher = None

    def on_close(self, cancelable=False):
        """Perform actions before parent main window is closed"""
        self.get_widget().save_config()
        self.get_widget().watcher.stop()
        return True

    def on_mainwindow_visible(self):
        # Open project passed on the command line or reopen last one.
        cli_options = self.get_command_line_options()
        initial_cwd = self._main.get_initial_working_directory()

        # There's no need to restart the console if the user wants to connect
        # to a kernel at startup.
        # Fixes spyder-ide/spyder#23497
        restart_console = cli_options.connection_file is None

        if cli_options.project is not None:
            logger.debug('Opening project from the command line')
            project = osp.normpath(
                osp.join(initial_cwd, cli_options.project)
            )
            self.open_project(
                project,
                workdir=cli_options.working_directory,
                restart_console=restart_console
            )
        else:
            self.get_widget().set_pane_empty()
            logger.debug('Reopening project from last session')
            self.get_widget().reopen_last_project(
                working_directory=cli_options.working_directory,
                restart_console=restart_console
            )

    # ---- Public API
    # -------------------------------------------------------------------------
    def create_project(self, path, project_type_id=EmptyProject.ID):
        """
        Create a new project.

        Parameters
        ----------
        path: str
            Filesystem path where the project will be created.
        project_type_id: str, optional
            Id for the project type. The default is 'empty-project-type'.
        packages: list, optional
            Package to install. Currently not in use.
        """
        self.get_widget().create_project(path, project_type_id)

    def open_project(self, path=None, project_type=None, restart_console=True,
                     save_previous_files=True, workdir=None):
        """
        Open the project located in a given path.

        Parameters
        ----------
        path: str
            Filesystem path where the project is located.
        project_type: spyder.plugins.projects.api.BaseProjectType, optional
            Project type class.
        restart_console: bool, optional
            Whether to restart the IPython console (i.e. close all consoles and
            reopen a single one) after opening the project. The default is
            True.
        save_previous_files: bool, optional
            Whether to save the list of previous open files in the editor
            before opening the project. The default is True.
        workdir: str, optional
            Working directory to set after opening the project. The default is
            None.
        """
        self.get_widget().open_project(
            path, project_type, restart_console, save_previous_files, workdir
        )

    def close_project(self):
        """
        Close current project and return to a window without an active project.
        """
        self.get_widget().close_project()

    def delete_project(self):
        """
        Delete the current project without deleting the files in the directory.
        """
        self.get_widget().delete_project()

    def get_active_project(self):
        """Get the active project."""
        return self.get_widget().current_active_project

    def get_project_filenames(self):
        """Get the list of recent filenames of a project."""
        return self.get_widget().get_project_filenames()

    def set_project_filenames(self, filenames):
        """
        Set the list of open file names in a project.

        Parameters
        ----------
        filenames: list of strings
            File names to save in the project config options.
        """
        self.get_widget().set_project_filenames(filenames)

    def get_active_project_path(self):
        """Get path of the active project."""
        return self.get_widget().get_active_project_path()

    def get_last_working_dir(self):
        """Get the path of the last working directory."""
        return self.get_conf(
            'last_working_dir', section='editor', default=getcwd_or_home()
        )

    def is_valid_project(self, path):
        """
        Check if a directory is a valid Spyder project.

        Parameters
        ----------
        path: str
            Filesystem path to the project.
        """
        return self.get_widget().is_valid_project(path)

    def start_workspace_services(self):
        """Enable LSP workspace functionality."""
        self.get_widget().start_workspace_services()

    def stop_workspace_services(self, _language):
        """Disable LSP workspace functionality."""
        self.get_widget().stop_workspace_services()

    def register_project_type(self, parent_plugin, project_type):
        """
        Register a new project type.

        Parameters
        ----------
        parent_plugin: spyder.plugins.api.plugins.SpyderPluginV2
            The parent plugin instance making the project type registration.
        project_type: spyder.plugins.projects.api.BaseProjectType
            Project type to register.
        """
        self.get_widget().register_project_type(parent_plugin, project_type)

    def get_project_types(self):
        """
        Return available registered project types.

        Returns
        -------
        dict
            Project types dictionary. Keys are project type IDs and values
            are project type classes.
        """
        return self.get_widget().get_project_types()

    # ---- Private API
    # -------------------------------------------------------------------------
    def _new_editor(self, text):
        editor = self.get_plugin(Plugins.Editor)
        editor.new(text=text)

    def _setup_editor_files(self, __unused):
        editor = self.get_plugin(Plugins.Editor)
        editor.setup_open_files()

    def _set_path_in_editor(self, path):
        editor = self.get_plugin(Plugins.Editor)
        editor.set_current_project_path(path)

    def _unset_path_in_editor(self, __unused):
        editor = self.get_plugin(Plugins.Editor)
        editor.set_current_project_path()

    def _add_path_to_completions(self, path):
        self._completions.project_path_update(
            path,
            update_kind=WorkspaceUpdateKind.ADDITION,
            instance=self.get_widget()
        )

    def _remove_path_from_completions(self, path):
        self._completions.project_path_update(
            path,
            update_kind=WorkspaceUpdateKind.DELETION,
            instance=self.get_widget()
        )

    def _run_file_in_ipyconsole(self, fname):
        ipyconsole = self.get_plugin(Plugins.IPythonConsole)
        ipyconsole.run_script(
            filename=fname,
            wdir=osp.dirname(fname),
            current_client=False,
            clear_variables=True
        )

    def _show_main_widget(self):
        """Show the main widget."""
        if self.get_widget() is not None:
            self.get_widget().show_widget()

    def _get_open_filenames(self):
        editor = self.get_plugin(Plugins.Editor)
        if editor is not None:
            return editor.get_filenames()
        return []

    def _is_invalid_active_project(self):
        """Handle an invalid active project."""
        self.get_widget().is_invalid_active_project()

    def _broadcast_notification(self, method, params):
        self._completions.broadcast_notification(method, params)

    def _handle_switcher_modes(self, mode):
        """
        Populate switcher with files in active project.

        List the file names of the current active project with their
        directories in the switcher. It only handles the files mode, i.e.
        an empty string.

        Parameters
        ----------
        mode: str
            The selected mode (open files "", symbol "@" or line ":").
        """
        # Don't compute anything if we're not in files mode
        if mode != "":
            return

        self.get_widget().display_default_switcher_items()

    def _handle_switcher_selection(self, item, mode, search_text):
        """
        Handle user selecting item in switcher.

        If the selected item is not in the section of the switcher that
        corresponds to this plugin, then ignore it. Otherwise, switch to
        selected project file and hide the switcher.

        Parameters
        ----------
        item: object
            The current selected item from the switcher list (QStandardItem).
        mode: str
            The current selected mode (open files "", symbol "@" or line ":").
        search_text: str
            Cleaned search/filter text.
        """
        self.get_widget().handle_switcher_selection(item, mode, search_text)
        self._switcher.hide()

    def _handle_switcher_search(self, search_text):
        """
        Handle user typing in switcher to filter results.

        Parameters
        ----------
        text: str
            The current search text in the switcher dialog box.
        """
        self.get_widget().handle_switcher_search(search_text)

    def _display_items_in_switcher(self, items, setup, clear_section):
        """
        Display a list of items in the switcher.

        Parameters
        ----------
        items: list
            Items to display.
        setup: bool
            Call the switcher's setup after adding the items.
        clear_section: bool
            Clear Projects section before adding the items.
        """
        if clear_section:
            self._switcher.remove_section(self.get_widget().get_title())

        for (title, description, icon, section, path, is_last_item) in items:
            self._switcher.add_item(
                title=title,
                description=description,
                icon=icon,
                section=section,
                data=path,
                last_item=is_last_item,
                score=1e10,  # To make the editor results appear first
                use_score=False  # Results come from fzf in the right order
            )

        if setup:
            self._switcher.setup()



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
#
"""Tests for workingdirectory plugin."""

# Standard library imports
import os
import os.path as osp
import sys
from unittest.mock import Mock

# Third-party imports
import pytest
from qtpy.QtWidgets import QMainWindow
from qtpy.QtCore import Qt

# Local imports
from spyder.app.cli_options import get_options
from spyder.config.base import get_home_dir
from spyder.config.manager import CONF
from spyder.plugins.workingdirectory.plugin import WorkingDirectory


NEW_DIR = 'new_workingdir'


class MainWindow(QMainWindow):

    def __init__(self):
        # This avoids using the cli options passed to pytest
        sys_argv = [sys.argv[0]]
        self._cli_options = get_options(sys_argv)[0]
        super().__init__()

    def get_plugin(self, plugin, error=True):
        return Mock()


@pytest.fixture
def setup_workingdirectory(qtbot, request, tmpdir):
    """Setup working directory plugin."""
    use_startup_wdir = request.node.get_closest_marker('use_startup_wdir')
    use_cli_wdir = request.node.get_closest_marker('use_cli_wdir')

    # Setting default options
    CONF.set('workingdir', 'startup/use_project_or_home_directory', True)
    CONF.set('workingdir', 'startup/use_fixed_directory', False)

    # Create main window and new directory
    main_window = MainWindow()

    if use_startup_wdir:
        new_wdir = tmpdir.mkdir(NEW_DIR + '_startup')
        CONF.set('workingdir', 'startup/use_project_or_home_directory', False)
        CONF.set('workingdir', 'startup/use_fixed_directory', True)
        CONF.set('workingdir', 'startup/fixed_directory', str(new_wdir))
    elif use_cli_wdir:
        new_wdir = tmpdir.mkdir(NEW_DIR + '_cli')
        main_window._cli_options.working_directory = str(new_wdir)

    workingdirectory = WorkingDirectory(main_window, configuration=CONF)
    workingdirectory.on_initialize()
    workingdirectory.close = lambda: True

    return workingdirectory


def test_basic_initialization(setup_workingdirectory):
    """Test Working Directory plugin initialization."""
    workingdirectory = setup_workingdirectory

    # Assert that workingdirectory exists
    assert workingdirectory is not None


def test_get_workingdir(setup_workingdirectory):
    """Test the method that defines the working directory at home."""
    workingdirectory = setup_workingdirectory
    # Start the working directory on the home directory
    act_wdir = workingdirectory.get_workdir()
    assert act_wdir == get_home_dir()


@pytest.mark.use_startup_wdir
def test_get_workingdir_startup(setup_workingdirectory):
    """
    Test the method that sets the working directory according to the one
    selected in preferences.
    """
    workingdirectory = setup_workingdirectory

    # Get the current working directory
    cwd = workingdirectory.get_workdir()
    folders = osp.split(cwd)

    # Asert working directory is the expected one
    assert folders[-1] == NEW_DIR + '_startup'


@pytest.mark.use_cli_wdir
def test_get_workingdir_cli(setup_workingdirectory):
    """
    Test that the plugin sets the working directory passed by users on the
    command line with the --workdir option.
    """
    workingdirectory = setup_workingdirectory

    # Get the current working directory
    cwd = workingdirectory.get_container().history[-1]
    folders = osp.split(cwd)

    # Asert working directory is the expected one
    assert folders[-1] == NEW_DIR + '_cli'


def test_file_goto(qtbot, setup_workingdirectory):
    """
    Test that putting a file in the workingdirectory emits a edit_goto signal.
    """
    container = setup_workingdirectory.get_container()
    
    signal_res = {}
    
    def test_slot(filename, line, word):
        signal_res["filename"] = filename
        signal_res["line"] = line
    
    container.edit_goto.connect(test_slot)
    
    pathedit = container.pathedit
    wd = setup_workingdirectory.get_workdir()
    filename = osp.join(wd, "myfile_workingdirectory_test.py")
    with open(filename, "w") as f:
        f.write("\n" * 5)
    with qtbot.waitSignal(container.edit_goto):
        pathedit.add_text(filename + ":1")
        qtbot.keyClick(pathedit, Qt.Key_Return)
    
    assert signal_res["filename"] in filename
    assert signal_res["line"] == 1
    
    os.remove(filename)


if __name__ == "__main__":
    pytest.main()



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""
Main menu Plugin.
"""

# Standard library imports
from collections import OrderedDict
import os
import sys
from typing import Dict, List, Tuple, Optional, Union

# Local imports
from spyder.api.exceptions import SpyderAPIError
from spyder.api.fonts import SpyderFontType
from spyder.api.plugin_registration.registry import PLUGIN_REGISTRY
from spyder.api.plugins import SpyderPluginV2, SpyderDockablePlugin, Plugins
from spyder.api.translations import _
from spyder.api.widgets.menus import SpyderMenu
from spyder.api.widgets.mixins import SpyderMenuMixin
from spyder.plugins.mainmenu.api import (
    ApplicationMenu,
    ApplicationMenus,
    MENUBAR_STYLESHEET,
)
from spyder.utils.qthelpers import SpyderAction


# Extended typing definitions
ItemType = Union[SpyderAction, SpyderMenu]
ItemSectionBefore = Tuple[
    ItemType, Optional[str], Optional[str], Optional[str]]
ItemQueue = Dict[str, List[ItemSectionBefore]]


class MainMenu(SpyderPluginV2, SpyderMenuMixin):
    NAME = 'mainmenu'
    CONF_SECTION = NAME
    CONF_FILE = False
    CAN_BE_DISABLED = False

    @staticmethod
    def get_name():
        return _('Main menus')

    @classmethod
    def get_icon(cls):
        return cls.create_icon('genprefs')

    @staticmethod
    def get_description():
        return _('Provide main application menu management.')

    def on_initialize(self):
        # Reference holder dict for the menus
        self._APPLICATION_MENUS = OrderedDict()

        # Queue that contain items that are pending to add to a non-existing
        # menu
        self._ITEM_QUEUE = {}  # type: ItemQueue

        # Set style. This is only necessary on Windows and Linux
        if not sys.platform == 'darwin':
            app_font = self.get_font(font_type=SpyderFontType.Interface)
            self.main.menuBar().setFont(app_font)
            self.main.menuBar().setStyleSheet(str(MENUBAR_STYLESHEET))

        # Create Application menus using plugin public API
        create_app_menu = self.create_application_menu
        create_app_menu(ApplicationMenus.File, _("&File"))
        create_app_menu(ApplicationMenus.Edit, _("&Edit"))
        create_app_menu(ApplicationMenus.Search, _("&Search"))
        create_app_menu(ApplicationMenus.Source, _("Sour&ce"))
        create_app_menu(ApplicationMenus.Run, _("&Run"))
        create_app_menu(ApplicationMenus.Debug, _("&Debug"))
        if self.is_plugin_enabled(Plugins.IPythonConsole):
            create_app_menu(ApplicationMenus.Consoles, _("C&onsoles"))
        if self.is_plugin_enabled(Plugins.Projects):
            create_app_menu(
                ApplicationMenus.Projects,
                _("&Projects"),
                min_width=150 if os.name == "nt" else 170
            )
        create_app_menu(ApplicationMenus.Tools, _("&Tools"))
        create_app_menu(ApplicationMenus.View, _("&View"))
        create_app_menu(ApplicationMenus.Help, _("&Help"))

    def on_mainwindow_visible(self):
        # Pre-render menus so actions with menu roles (like "About Spyder" and
        # "Preferences") are located in the right place in Mac's menu bar.
        # Fixes spyder-ide/spyder#14917
        # This also registers shortcuts for actions that are only in menus.
        # Fixes spyder-ide/spyder#16061
        for menu in self._APPLICATION_MENUS.values():
            menu.render()

    # ---- Private methods
    # ------------------------------------------------------------------------
    def _hide_options_menus(self):
        """Hide options menu when menubar is pressed in macOS."""
        for plugin_name in PLUGIN_REGISTRY:
            plugin_instance = PLUGIN_REGISTRY.get_plugin(plugin_name)
            if isinstance(plugin_instance, SpyderDockablePlugin):
                if plugin_instance.CONF_SECTION == 'editor':
                    editorstack = self._main.editor.get_current_editorstack()
                    editorstack.menu.hide()
                else:
                    try:
                        # New API
                        plugin_instance.options_menu.hide()
                    except AttributeError:
                        # Old API
                        plugin_instance._options_menu.hide()

    # ---- Public API
    # ------------------------------------------------------------------------
    def create_application_menu(
        self,
        menu_id: str,
        title: str,
        min_width: Optional[int] = None
    ):
        """
        Create a Spyder application menu.

        Parameters
        ----------
        menu_id: str
            The menu unique identifier string.
        title: str
            The localized menu title to be displayed.
        min_width: int
            Minimum width for the menu in pixels.
        """
        if menu_id in self._APPLICATION_MENUS:
            raise SpyderAPIError(
                'Menu with id "{}" already added!'.format(menu_id)
            )

        menu = self._create_menu(
            menu_id=menu_id,
            parent=self.main,
            title=title,
            min_width=min_width,
            MenuClass=ApplicationMenu
        )
        self._APPLICATION_MENUS[menu_id] = menu
        self.main.menuBar().addMenu(menu)

        if sys.platform == 'darwin':
            menu.aboutToShow.connect(self._hide_options_menus)

            # This is necessary because for some strange reason the
            # "Configuration per file" entry disappears after showing other
            # dialogs and the only way to make it visible again is by
            # re-rendering the menu.
            if menu_id == ApplicationMenus.Run:
                menu.aboutToShow.connect(lambda: menu.render(force=True))

        if menu_id in self._ITEM_QUEUE:
            pending_items = self._ITEM_QUEUE.pop(menu_id)
            for pending in pending_items:
                (item, section,
                 before_item, before_section) = pending
                self.add_item_to_application_menu(
                    item, menu_id=menu_id, section=section,
                    before=before_item, before_section=before_section)

        return menu

    def add_item_to_application_menu(self, item: ItemType,
                                     menu_id: Optional[str] = None,
                                     section: Optional[str] = None,
                                     before: Optional[str] = None,
                                     before_section: Optional[str] = None,
                                     omit_id: bool = False):
        """
        Add action or widget `item` to given application menu `section`.

        Parameters
        ----------
        item: SpyderAction or SpyderMenu
            The item to add to the `menu`.
        menu_id: str or None
            The application menu unique string identifier.
        section: str or None
            The section id in which to insert the `item` on the `menu`.
        before: str
            Make the item appear before the given object identifier.
        before_section: Section or None
            Make the item section (if provided) appear before another
            given section.
        omit_id: bool
            If True, then the menu will check if the item to add declares an
            id, False otherwise. This flag exists only for items added on
            Spyder 4 plugins. Default: False

        Notes
        -----
        Must provide a `menu` or a `menu_id`.
        """
        if not isinstance(item, (SpyderAction, SpyderMenu)) and not omit_id:
            raise SpyderAPIError('A menu only accepts items objects of type '
                                 'SpyderAction or SpyderMenu')

        if menu_id not in self._APPLICATION_MENUS:
            pending_menu_items = self._ITEM_QUEUE.get(menu_id, [])
            pending_menu_items.append((item, section, before,
                                       before_section))
            self._ITEM_QUEUE[menu_id] = pending_menu_items
        else:
            menu = self.get_application_menu(menu_id)
            menu.add_action(item, section=section, before=before,
                            before_section=before_section, omit_id=omit_id)

    def remove_application_menu(self, menu_id: str):
        """
        Remove a Spyder application menu.

        Parameters
        ----------
        menu_id: str
            The menu unique identifier string.
        """
        if menu_id in self._APPLICATION_MENUS:
            menu = self._APPLICATION_MENUS.pop(menu_id)
            self.main.menuBar().removeAction(menu.menuAction())

    def remove_item_from_application_menu(self, item_id: str,
                                          menu_id: Optional[str] = None):
        """
        Remove action or widget from given application menu by id.

        Parameters
        ----------
        item_id: str
            The item identifier to remove from the given menu.
        menu_id: str or None
            The application menu unique string identifier.
        """
        if menu_id not in self._APPLICATION_MENUS:
            raise SpyderAPIError('{} is not a valid menu_id'.format(menu_id))

        menu = self.get_application_menu(menu_id)
        menu.remove_action(item_id)

    def get_application_menu(self, menu_id: str) -> SpyderMenu:
        """
        Return an application menu by menu unique id.

        Parameters
        ----------
        menu_id: ApplicationMenu
            The menu unique identifier string.
        """
        if menu_id not in self._APPLICATION_MENUS:
            raise SpyderAPIError(
                'Application menu "{0}" not found! Available '
                'menus are: {1}'.format(
                    menu_id, list(self._APPLICATION_MENUS.keys()))
            )

        return self._APPLICATION_MENUS[menu_id]



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""Tests for config/manager.py."""

# Standard library imports
import configparser
import os
import os.path as osp
import shutil
import sys

# Third party imports
import keyring
import pytest

# Local imports
from spyder.config.base import get_conf_path, get_conf_paths, running_in_ci
from spyder.config.manager import ConfigurationManager
from spyder.plugins.console.plugin import Console


def clear_site_config():
    """Delete all test site config folders."""
    for path in get_conf_paths():
        shutil.rmtree(path)


def test_site_config_load():
    """
    Test that the site/system config preferences are loaded with correct
    precedence.
    """
    clear_site_config()

    for i, path in enumerate(reversed(get_conf_paths())):
        exp_value = 100*(1 + i)
        content = '[main]\nmemory_usage/timeout = ' + str(exp_value) + '\n'

        conf_fpath = os.path.join(path, 'spyder.ini')
        with open(conf_fpath, 'w') as fh:
            fh.write(content)

        config = ConfigurationManager()
        config.reset_to_defaults()
        value = config.get('main', 'memory_usage/timeout')

        print(path, value, exp_value)
        assert value == exp_value

    clear_site_config()


def test_external_plugin_config(qtbot):
    """
    Test that config for external plugins is saved as expected.

    This includes a regression for part two (the shortcuts conflict) of
    issue spyder-ide/spyder#11132
    """
    clear_site_config()

    # Emulate an old Spyder 3 configuration
    spy3_config = """
[main]
version = 1.0.0

[shortcuts]
foo/bar = Alt+1

[ipython_console]
startup/run_lines =
"""
    conf_fpath = get_conf_path('spyder.ini')
    with open(conf_fpath, 'w') as f:
        f.write(spy3_config)

    # Create config manager
    manager = ConfigurationManager()

    # Set default options for the internal console
    Console.CONF_FILE = True
    defaults = [
        ('internal_console',
         {
            'max_line_count': 300,
            'working_dir_history': 30,
            'working_dir_adjusttocontents': False,
            'wrap': True,
            'codecompletion/auto': False,
            'external_editor/path': 'SciTE',
            'external_editor/gotoline': '-goto:'
         }
        ),
    ]
    Console.CONF_DEFAULTS = defaults

    # Register console plugin config
    manager.register_plugin(Console)

    # Assert the dummy shortcut is not present in the plugin config
    with pytest.raises(configparser.NoSectionError):
        manager.get_shortcut('foo', 'bar', plugin_name='internal_console')

    # Change an option in the console
    console = Console(None, configuration=manager)
    console.set_conf('max_line_count', 600)

    # Read config filew directly
    user_path = manager.get_user_config_path()
    with open(osp.join(user_path, 'spyder.ini'), 'r') as f:
        user_contents = f.read()

    plugin_path = manager.get_plugin_config_path('internal_console')
    with open(osp.join(plugin_path, 'spyder.ini'), 'r') as f:
        plugin_contents = f.read()

    # Assert that the change was written to the right config file
    assert 'max_line_count = 600' not in user_contents
    assert 'max_line_count = 600' in plugin_contents

    shutil.rmtree(plugin_path)
    Console.CONF_FILE = False
    clear_site_config()


@pytest.mark.skipif(
    sys.platform.startswith("linux") and running_in_ci(),
    reason="Fails on Linux and CIs",
)
def test_secure_options():
    """Test that we can save and retrieve secure options."""
    clear_site_config()

    # Set a secure option
    config = ConfigurationManager()
    config.set("test_section", "test_option", "secure-option", secure=True)

    # Check the option was not saved directly in our config system
    with pytest.raises(configparser.NoSectionError):
        config.get("test_section", "test_option", secure=False)

    # Check the option is retrieved correctly
    option = config.get("test_section", "test_option", secure=True)
    assert option == "secure-option"

    # Remove the option with keyring. If the option was not saved with it, this
    # will give an error.
    keyring.delete_password("test_section", "test_option")
    assert keyring.get_password("test_section", "test_option") is None

    clear_site_config()


if __name__ == "__main__":
    pytest.main()



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
#
'''
Tests for editor panels.
'''

# Third party imports
import pytest

# Local imports
from spyder.utils.qthelpers import qapplication
from spyder.plugins.editor.panels.linenumber import LineNumberArea
from spyder.plugins.editor.panels.edgeline import EdgeLine
from spyder.plugins.editor.panels.scrollflag import ScrollFlagArea
from spyder.plugins.editor.panels.indentationguides import IndentationGuide
from spyder.plugins.editor.widgets.codeeditor import CodeEditor


# --- Fixtures
# -----------------------------------------------------------------------------
def construct_editor(*args, **kwargs):
    app = qapplication()
    editor = CodeEditor(parent=None)
    kwargs['language'] = 'Python'
    editor.setup_editor(*args, **kwargs)
    return editor


# --- Tests
# -----------------------------------------------------------------------------
@pytest.mark.parametrize('state', [True, False])
@pytest.mark.parametrize('setting, panelclass', [
    ('linenumbers', LineNumberArea),
    ('edge_line', EdgeLine),
    ('scrollflagarea', ScrollFlagArea),
    ('indent_guides', IndentationGuide),
])
def test_activate_panels(setting, panelclass, state):
    """Test activate/deactivate of editors Panels.

    Also test that the panel is added to the editor.
    """
    kwargs = {}
    kwargs[setting] = state
    editor = construct_editor(**kwargs)

    found = False
    for panel in editor.panels:
        if isinstance(panel, panelclass):
            assert panel.enabled == state
            found = True
    assert found


if __name__ == '__main__':
    pytest.main()



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""Plugin registry configuration page."""

# Third party imports
from pyuca import Collator
from qtpy.QtWidgets import QVBoxLayout, QLabel

# Local imports
from spyder.api.preferences import PluginConfigPage
from spyder.config.base import _
from spyder.widgets.elementstable import ElementsTable


class PluginsConfigPage(PluginConfigPage):

    def setup_page(self):
        newcb = self.create_checkbox
        self.plugins_checkboxes = {}

        header_label = QLabel(
            _("Disable a Spyder plugin (external or built-in) to prevent it "
              "from loading until re-enabled here, to simplify the interface "
              "or in case it causes problems.")
        )
        header_label.setWordWrap(True)

        # To save the plugin elements
        internal_elements = []
        external_elements = []

        # ------------------ Internal plugins ---------------------------------
        for plugin_name in self.plugin.all_internal_plugins:
            (conf_section_name,
             PluginClass) = self.plugin.all_internal_plugins[plugin_name]

            if not getattr(PluginClass, 'CAN_BE_DISABLED', True):
                # Do not list core plugins that can not be disabled
                continue

            plugin_state = self.get_option(
                'enable', section=conf_section_name, default=True)
            cb = newcb('', 'enable', default=True, section=conf_section_name,
                       restart=True)

            internal_elements.append(
                dict(
                    title=PluginClass.get_name(),
                    description=PluginClass.get_description(),
                    icon=PluginClass.get_icon(),
                    widget=cb,
                    additional_info=_("Built-in")
                )
            )

            self.plugins_checkboxes[plugin_name] = (cb.checkbox, plugin_state)

        # ------------------ External plugins ---------------------------------
        for plugin_name in self.plugin.all_external_plugins:
            (conf_section_name,
             PluginClass) = self.plugin.all_external_plugins[plugin_name]

            if not getattr(PluginClass, 'CAN_BE_DISABLED', True):
                # Do not list external plugins that can not be disabled
                continue

            plugin_state = self.get_option(
                f'{conf_section_name}/enable',
                section=self.plugin._external_plugins_conf_section,
                default=True
            )
            cb = newcb('', f'{conf_section_name}/enable', default=True,
                       section=self.plugin._external_plugins_conf_section,
                       restart=True)

            external_elements.append(
                dict(
                    title=PluginClass.get_name(),
                    description=PluginClass.get_description(),
                    icon=PluginClass.get_icon(),
                    widget=cb
                )
            )

            self.plugins_checkboxes[plugin_name] = (cb.checkbox, plugin_state)

        # Sort elements by title for easy searching
        collator = Collator()
        internal_elements.sort(key=lambda e: collator.sort_key(e['title']))
        external_elements.sort(key=lambda e: collator.sort_key(e['title']))

        # Build plugins table, showing external plugins first.
        plugins_table = ElementsTable(
            self, external_elements + internal_elements
        )

        # Layout
        layout = QVBoxLayout()
        layout.addWidget(header_label)
        layout.addSpacing(15)
        layout.addWidget(plugins_table)
        layout.addSpacing(15)
        self.setLayout(layout)

    def apply_settings(self):
        for plugin_name in self.plugins_checkboxes:
            cb, previous_state = self.plugins_checkboxes[plugin_name]
            if cb.isChecked() and not previous_state:
                self.plugin.set_plugin_enabled(plugin_name)
                PluginClass = None
                external = False
                if plugin_name in self.plugin.all_internal_plugins:
                    (__,
                     PluginClass) = self.plugin.all_internal_plugins[plugin_name]
                elif plugin_name in self.plugin.all_external_plugins:
                    (__,
                     PluginClass) = self.plugin.all_external_plugins[plugin_name]
                    external = True  # noqa

                # TODO: Once we can test that all plugins can be restarted
                # without problems during runtime, we can enable the
                # autorestart feature provided by the plugin registry:
                # self.plugin.register_plugin(self.main, PluginClass,
                #                             external=external)
            elif not cb.isChecked() and previous_state:
                # TODO: Once we can test that all plugins can be restarted
                # without problems during runtime, we can enable the
                # autorestart feature provided by the plugin registry:
                # self.plugin.delete_plugin(plugin_name)
                pass
        return set({})



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""External terminal API."""

# Standard library imports
from __future__ import annotations
from typing import TypedDict


class ExtTerminalPyConfiguration(TypedDict):
    """External terminal execution parameters for Python files."""

    # True if the external terminal is using custom arguments. False otherwise
    args_enabled: bool

    # Custom arguments to pass to the external terminal.
    args: str

    # True if the terminal should remain open once the execution finishes.
    # False otherwise.
    interact: bool

    # True if the terminal is using custom Python arguments. False otherwise.
    python_args_enabled: bool

    # Custom arguments to pass to the terminal.
    python_args: str


class ExtTerminalShConfiguration(TypedDict):
    """External terminal execution parameters for shell files."""

    # Path/name of the shell interpreter to use.
    interpreter: str

    # True if the shell interpreter is using custom arguments, False otherwise.
    interpreter_opts_enabled: bool

    # Custom arguments to pass to the shell interpreter.
    interpreter_opts: str

    # True if the shell script is using custom arguments, False otherwise.
    script_opts_enabled: bool

    # Custom arguments to pass to the shell script.
    script_opts: str

    # True if the terminal will be closed after execution, False otherwise.
    close_after_exec: bool
